<?php

namespace Drupal\activitypub\Commands;

use Drush\Commands\DrushCommands;

/**
 * ActivityPub Drush commands file.
 */
class ActivityPubCommands extends DrushCommands {

  /**
   * Process outbox.
   *
   * @param int $send
   *   Whether to send to activities to inboxes.
   * @param int $remove_queue_item
   *   Delete item from queue.
   * @param int $debug
   *   Whether to view debug statements.
   *
   * @command activitypub:process-outbox
   */
  public function processOutbox($send = 1, $remove_queue_item = 1, $debug = 0) {
    if (\Drupal::config('activitypub.settings')->get('process_outbox_handler') == 'drush') {

      // Make sure the host is not set to default.
      if (\Drupal::request()->getHost() == 'default') {
        $this->writeln("Please provide the -l or --uri parameter to set the host.");
        return;
      }

      \Drupal::service('activitypub.process.client')->handleOutboxQueue((bool) $send, (bool) $remove_queue_item, (bool) $debug);
    }
  }

  /**
   * Process inbox.
   *
   * @param int $remove_queue_item
   *   Delete item from queue.
   * @param int $debug
   *   Whether to view debug statements.
   *
   * @command activitypub:process-inbox
   */
  public function processInbox($remove_queue_item = 1, $debug = 0) {
    if (\Drupal::config('activitypub.settings')->get('process_inbox_handler') == 'drush') {

      // Make sure the host is not set to default.
      if (\Drupal::request()->getHost() == 'default') {
        $this->writeln("Please provide the -l or --uri parameter to set the host.");
        return;
      }

      \Drupal::service('activitypub.process.client')->handleInboxQueue((bool) $remove_queue_item, (bool) $debug);
    }
  }

  /**
   * Add activity to queue.
   *
   * @param $activityId
   *
   * @command activitypub:add-to-queue
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function addActivityToQueue($activityId) {
    /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity */
    $activity = \Drupal::entityTypeManager()->getStorage('activitypub_activity')->load($activityId);
    if ($activity) {
      $activity
        ->set('processed', FALSE)
        ->set('queued', TRUE)
        ->save();
      \Drupal::service('activitypub.process.client')->createQueueItem($activity);
      $this->logger()->notice('Added to queue');
    }
    else {
      $this->logger()->notice('Activity not found');
    }
  }

  /**
   * Fetches and caches remote users for a user.
   *
   * @param $uid
   *
   * @command activitypub:fetch-remote-users
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function cacheRemoteUsers($uid) {
    $server = \Drupal::service('activitypub.utility')->getServer();
    $conditions = ['type' => 'Follow', 'uid' => $uid, 'status' => 1];
    /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface[] $activities */
    $activities = \Drupal::entityTypeManager()->getStorage('activitypub_activity')->loadByProperties($conditions);
    foreach ($activities as $activity) {
      $this->writeln('Fetching ' . $activity->getActor());
      $server->actor($activity->getActor());
    }
  }

  /**
   * Get info about a name via WebFinger.
   *
   * @param $handle
   *   The handle to fetch.
   * @param int $debug
   *   Whether to debug or not.
   *
   * @command activitypub:webfinger-info
   */
  public function getWebfingerInfo($handle, $debug = 1) {
    /** @var \Drupal\activitypub\Services\ActivityPubUtilityInterface $utility */
    $utility = \Drupal::service('activitypub.utility');
    $server = $utility->getServer(['instance' => ['debug' => (bool) $debug]]);

    // Get actor.
    $actor = $server->actor($handle);
    print_r($actor->get('inbox'));

    // Get a WebFinger instance
    $webfinger = $actor->webfinger();
    print_r($webfinger->toArray());
  }
}
