<?php

namespace Drupal\activitypub\Controller;

use ActivityPhp\Server\Http\HttpSignature;
use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\activitypub\Entity\ActivitypubActorInterface;
use Drupal\user\UserInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

class InboxController extends BaseController {

  /**
   * Inbox routing callback.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   * @param \Drupal\user\UserInterface $user
   * @param \Drupal\activitypub\Entity\ActivitypubActorInterface $activitypub_actor
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   */
  public function inbox(Request $request, UserInterface $user, ActivitypubActorInterface $activitypub_actor) {
    $status = 400;
    $payload = @json_decode((string)$request->getContent(), TRUE);

    if ($request->getMethod() == 'POST' && !empty($payload)) {
      try {

        // The signature check is used to set the status of the activity. There
        // is a big chance some might fail depending how the request is signed
        // and which rfc version is used. In case the verification fails, we
        // allow posts to be followed in case the actor is a followee of the
        // current user.
        try {
          $httpSignature = new HttpSignature($this->activityPubUtility->getServer());
          $published = $httpSignature->verify($request);
          if (!$published && isset($payload['actor']) && isset($payload['type']) && in_array($payload['type'], $this->activityPubUtility->getTimelineTypes())) {
            $published = $this->isFollowee($payload['actor'], $activitypub_actor->getOwnerId());
            if ($published) {
              $this->getLogger('activitypub')->notice('Post published: @actor follows @followee', ['@actor' => $activitypub_actor->getName(), '@followee' => $payload['actor']]);
            }
            else {
              $this->getLogger('activitypub')->notice('Post left unpublished: @actor does not follow @followee', ['@actor' => $activitypub_actor->getName(), '@followee' => $payload['actor']]);
            }
          }
        }
        catch (\Exception $e) {
          $published = FALSE;
          $this->getLogger('activitypub')->error('Signature check exception: @message', ['@message' => $e->getMessage()]);
        }

        // Save when the actor is not blocked.
        if (!$this->domainIsBlocked($activitypub_actor->getBlockedDomains(), $payload['actor'])) {

          $object = '';
          if (isset($payload['object'])) {
            if (is_array($payload['object']) && isset($payload['object']['object'])) {
              if ($payload['type'] == 'Accept' && isset($payload['object']['actor'])) {
                $object = $payload['object']['actor'];
              }
              else {
                $object = $payload['object']['object'];
              }
            }
            elseif (is_array($payload['object']) && isset($payload['object']['inReplyTo'])) {
              $object = $payload['object']['inReplyTo'];
            }
            elseif (is_array($payload['object']) && isset($payload['object']['id'])) {
              $object = $payload['object']['id'];
            }
            elseif (is_string($payload['object'])) {
              $object = $payload['object'];
            }
          }

          /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActivityStorage $storage */
          $storage = $this->entityTypeManager()->getStorage('activitypub_activity');
          $values = [
            'uid' => $user->id(),
            'collection' => ActivityPubActivityInterface::INBOX,
            'external_id' => $payload['id'],
            'type' => $payload['type'],
            'actor' => $payload['actor'],
            'object' => $object,
            'payload' => (string) $request->getContent(),
            'status' => $published,
          ];

          /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity */
          $activity = $storage->create($values);
          $doSave = TRUE;
          $activity->preInboxSave($doSave);
          if ($doSave) {
            $activity->save();
          }

          $status = 202;
        }
        else {
          $status = 403;
        }

      }
      catch (\Exception $e) {
        $this->getLogger('activitypub')->notice('Inbox error: @message - @content', ['@message' => $e->getMessage(), '@content' => (string) $request->getContent()]);
      }
    }

    return new JsonResponse(NULL, $status);
  }

  /**
   * Returns if the actor is followed by the user.
   *
   * @param $followee
   * @param $uid
   *
   * @return bool
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function isFollowee($followee, $uid) {

    $conditions = [
      'type' => 'Follow',
      'status' => 1,
      'object' => $followee,
      'uid' => $uid,
      'collection' => ActivityPubActivityInterface::OUTBOX,
    ];

    /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActivityStorage $storage */
    $storage = $this->entityTypeManager()->getStorage('activitypub_activity');
    $count = (int) $storage->getActivityCount($conditions);

    return $count === 1;
  }

  /**
   * Returns whether the actor is blocked or not.
   *
   * @param $domains
   * @param $actor
   *
   * @return false
   */
  protected function domainIsBlocked($domains, $actor) {
    $blocked = FALSE;

    if (!empty($domains)) {
      $blocked = $this->pathMatcher->matchPath($actor, $domains);
    }

    return $blocked;
  }

}
