<?php

namespace Drupal\activitypub\Controller;

use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\activitypub\Entity\ActivityPubActorInterface;
use Drupal\activitypub\Services\ActivityPubKeysInterface;
use Drupal\activitypub\Services\ActivityPubUtilityInterface;
use Drupal\Core\Cache\CacheableJsonResponse;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Url;
use Drupal\user\UserInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;

class UserController extends BaseController {

  /**
   * The ActivityPub keys service.
   *
   * @var \Drupal\activitypub\Services\ActivityPubKeysInterface
   */
  protected $activityPubKeys;

  /**
   * The ActivityPub utility service.
   *
   * @var \Drupal\activitypub\Services\ActivityPubUtilityInterface
   */
  protected $activityPubUtility;

  /**
   * SelfController constructor
   *
   * @param \Drupal\activitypub\Services\ActivityPubKeysInterface|null $activitypub_keys
   *   The ActivityPub keys service.
   * @param \Drupal\activitypub\Services\ActivityPubUtilityInterface $activitypub_utility
   *   The ActivityPub utility service.
   */
  public function __construct(ActivityPubKeysInterface $activitypub_keys, ActivityPubUtilityInterface $activitypub_utility) {
    $this->activityPubKeys = $activitypub_keys;
    $this->activityPubUtility = $activitypub_utility;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('activitypub.keys'),
      $container->get('activitypub.utility')
    );
  }

  /**
   * Self routing callback.
   *
   * @param \Drupal\user\UserInterface $user
   * @param \Drupal\activitypub\Entity\ActivityPubActorInterface $activitypub_actor
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   */
  public function self(UserInterface $user, ActivityPubActorInterface $activitypub_actor) {

    $metadata = new CacheableMetadata();
    $canonical = Url::fromRoute('entity.user.canonical', ['user' => $user->id()], ['absolute' => TRUE])->toString(TRUE);
    $inbox = Url::fromRoute('activitypub.inbox', ['user' => $user->id(), 'activitypub_actor' => $activitypub_actor->getName()], ['absolute' => TRUE])->toString(TRUE);
    $outbox = Url::fromRoute('activitypub.outbox', ['user' => $user->id(), 'activitypub_actor' => $activitypub_actor->getName()], ['absolute' => TRUE])->toString(TRUE);
    $following = Url::fromRoute('activitypub.following', ['user' => $user->id(), 'activitypub_actor' => $activitypub_actor->getName()], ['absolute' => TRUE])->toString(TRUE);
    $followers = Url::fromRoute('activitypub.followers', ['user' => $user->id(), 'activitypub_actor' => $activitypub_actor->getName()], ['absolute' => TRUE])->toString(TRUE);
    $metadata->addCacheableDependency($canonical);
    $metadata->addCacheableDependency($inbox);
    $metadata->addCacheableDependency($outbox);
    $metadata->addCacheableDependency($following);
    $metadata->addCacheableDependency($followers);
    $metadata->addCacheTags(['user:' . $user->id()]);

    $data = [
      'type' => 'Person',
      '@context' => [ActivityPubActivityInterface::CONTEXT_URL, ActivityPubActivityInterface::SECURITY_URL],
      'name' => $activitypub_actor->getName(),
      'preferredUsername' => $activitypub_actor->getName(),
      'id' => $this->activityPubUtility->getActivityPubID($activitypub_actor),
      'url' => $canonical->getGeneratedUrl(),
      'inbox' => $inbox->getGeneratedUrl(),
      'outbox' => $outbox->getGeneratedUrl(),
      'following' => $following->getGeneratedUrl(),
      'followers' => $followers->getGeneratedUrl(),
      'publicKey' => [
        'owner' => $this->activityPubUtility->getActivityPubID($activitypub_actor),
        'id' => $this->activityPubUtility->getActivityPubID($activitypub_actor) . '#main-key',
        'publicKeyPem' => $this->activityPubKeys->getPublicKey($activitypub_actor->getName()),
      ],
    ];

    if ($image_url = $this->activityPubUtility->getActivityPubUserAvatar($user)) {
      $image = [
        'type' => 'Image',
        'name' => $activitypub_actor->getName(),
        'url' => $image_url,
      ];
      $data['icon'] = (object) $image;
    }

    if ($summary = $activitypub_actor->getSummary()) {
      $data['summary'] = $summary;
    }

    $response = new CacheableJsonResponse($data);
    $response->addCacheableDependency($metadata);
    $response->headers->set('Content-Type', 'application/activity+json');
    return $response;
  }

  /**
   * User self routing callback with activity+json accept header.
   *
   * @param \Drupal\user\UserInterface $user
   *
   * @return \Drupal\Core\Cache\CacheableJsonResponse|\Symfony\Component\HttpFoundation\JsonResponse
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function selfJson(UserInterface $user) {
    /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorageInterface $storage */
    $storage = $this->entityTypeManager()->getStorage('activitypub_actor');
    $actor = $storage->loadActorByEntityIdAndType($user->id(), 'person');
    if ($actor) {
      return $this->self($user, $actor);
    }
    else {
      return new JsonResponse(NULL, 404);
    }
  }

  /**
   * Activities routing callback.
   *
   * @param \Drupal\user\UserInterface $user
   *
   * @return array
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function activities(UserInterface $user) {
    /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorageInterface $storage */
    $storage = $this->entityTypeManager()->getStorage('activitypub_actor');
    $actor = $storage->loadActorByEntityIdAndType($user->id(), 'person');
    if (!$actor) {

      $settings_link = Url::fromRoute('activitypub.user.settings', ['user' => $user->id()])->toString();

      return [
        'info' => ['#markup' => '<p>' . $this->t('ActivityPub is not enabled for your account.') . '</p>'],
        'link' => ['#markup' => '<p>' . $this->t('<a href="@url">Enable ActivityPub</a>', ['@url' => $settings_link]) . '</p>'],
      ];
    }
    else {
      return $this->entityTypeManager()->getListBuilder('activitypub_activity')->render();
    }
  }
}
