<?php

namespace Drupal\activitypub\Entity;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\user\EntityOwnerInterface;

/**
 * Provides an interface defining an activitypub activity entity type.
 */
interface ActivityPubActivityInterface extends ContentEntityInterface, EntityOwnerInterface {

  const FOLLOWERS = 'followers';
  const FOLLOWING = 'following';
  const INBOX = 'inbox';
  const OUTBOX = 'outbox';
  const CONTEXT_URL = 'https://www.w3.org/ns/activitystreams';
  const SECURITY_URL = 'https://w3id.org/security/v1';
  const PUBLIC_URL = 'https://www.w3.org/ns/activitystreams#Public';

  /**
   * Returns the object plugin manager.
   *
   * @return \Drupal\activitypub\Services\Type\TypePluginManager
   */
  public function getTypePluginManager();

  /**
   * Gets the activity creation timestamp.
   *
   * @return int
   *   Creation timestamp of the activity.
   */
  public function getCreatedTime();

  /**
   * Returns whether the activity is published or not.
   *
   * @return boolean
   */
  public function isPublished();

  /**
   * Returns whether the activity is processed or not.
   *
   * @return boolean
   */
  public function isProcessed();

  /**
   * Returns whether the activity is queued or not.
   *
   * @return boolean
   */
  public function isQueued();

  /**
   * Returns the collection.
   *
   * @return string
   */
  public function getCollection();

  /**
   * Returns the external ID.
   *
   * @return string
   */
  public function getExternalId();

  /**
   * Returns the config id.
   *
   * @return string
   */
  public function getConfigID();

  /**
   * Returns the Activity type.
   *
   * @return string
   */
  public function getType();

  /**
   * Returns the object.
   *
   * @return string
   */
  public function getObject();

  /**
   * Returns the Activity actor.
   *
   * @return string
   */
  public function getActor();

  /**
   * Returns the target Entity Type.
   *
   * @return string
   */
  public function getTargetEntityTypeId();

  /**
   * Return the target entity id.
   *
   * @return integer
   */
  public function getTargetEntityId();

  /**
   * Get the payload.
   *
   * @return string
   */
  public function getPayLoad();

  /**
   * Get the context.
   *
   * @return string
   */
  public function getContext();

  /**
   * Get array of URL's for the to property.
   *
   * @return array
   */
  public function getTo();

  /**
   * Returns whether the item was read or not.
   *
   * @return bool
   */
  public function isRead();

  /**
   * Get the raw to value.
   *
   * @return string
   */
  public function getToRaw();

  /**
   * Build the activity.
   *
   * @return array
   */
  public function buildActivity();

  /**
   * Inbox pre save call.
   *
   * @param $doSave
   */
  public function preInboxSave(&$doSave);

  /**
   * Act on entity delete.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity that is being deleted. It is a content entity, apart from the
   *   user or activitypub actor.
   */
  public function onEntityDelete(EntityInterface $entity);

  /**
   * Do inbox processing.
   */
  public function doInboxProcess();

  /**
   * Returns whether the activity can be queued.
   *
   * @return bool
   */
  public function canBeQueued();

  /**
   * Returns whether the activity can be undone.
   *
   * @return bool
   */
  public function canBeUndone();

}
