<?php

namespace Drupal\activitypub\Form;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

class ActivityPubSettingsForm extends ConfigFormBase {

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The entity field type plugin manager.
   *
   * @var \Drupal\Core\Field\FieldTypePluginManagerInterface
   */
  protected $fieldTypePluginManager;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * SettingsForm constructor
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   *   The entity field manager.
   * @param \Drupal\Core\Field\FieldTypePluginManagerInterface $fieldTypePluginManager
   *   The entity field type plugin manager.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   */
  public function __construct(ConfigFactoryInterface $config_factory, EntityFieldManagerInterface $entityFieldManager, FieldTypePluginManagerInterface $fieldTypePluginManager, ModuleHandlerInterface $module_handler) {
    parent::__construct($config_factory);
    $this->entityFieldManager = $entityFieldManager;
    $this->fieldTypePluginManager = $fieldTypePluginManager;
    $this->moduleHandler = $module_handler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_field.manager'),
      $container->get('plugin.manager.field.field_type'),
      $container->get('module_handler')
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['activitypub.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'activitypub_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $config = $this->config('activitypub.settings');

    $form['media'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Media'),
    ];

    $form['media']['avatar_user_style'] = [
      '#title' => $this->t('Avatar image style'),
      '#description' => $this->t('This image style will be used for the avatar for local and remote users.'),
      '#default_value' => $config->get('avatar_user_style'),
      '#type' => 'select',
      '#options' => image_style_options(FALSE),
    ];

    $user_image_fields = ['' => $this->t('- None -')];
    $fields = $this->entityFieldManager->getFieldStorageDefinitions('user');
    $field_types = $this->fieldTypePluginManager->getDefinitions();
    foreach ($fields as $key => $field) {
      if (in_array($field->getType(), ['image'])) {
        $user_image_fields[$key] = $field_types[$field->getType()]['label'] . ': ' . $field->getName();
      }
    }

    $form['media']['avatar_user_field'] = [
      '#title' => $this->t('Avatar field'),
      '#description' => $this->t('Select the field on the user which contains the user avatar.'),
      '#default_value' => $config->get('avatar_user_field'),
      '#type' => 'select',
      '#options' => $user_image_fields,
    ];

    $form['media']['attachment_content_style'] = [
      '#title' => $this->t('Attachment content image style'),
      '#description' => $this->t('This image style will be used for the image attachments on content.'),
      '#default_value' => $config->get('attachment_content_style'),
      '#type' => 'select',
      '#options' => image_style_options(TRUE),
    ];

    $form['media']['cache_images'] = [
      '#title' => $this->t('Cache images'),
      '#type' => 'checkbox',
      '#default_value' => $config->get('cache_images'),
      '#description' => $this->t('Cache avatars and photos in the Inbox locally using the imagecache external module.<br />Used by the Personal reader.'),
      '#disabled' => !$this->moduleHandler->moduleExists('imagecache_external'),
    ];

    $form['media']['cache_avatar_style'] = [
      '#title' => $this->t('Cache avatar style'),
      '#description' => $this->t('This image style will be used for the avatars in the inbox.'),
      '#default_value' => $config->get('cache_avatar_style'),
      '#type' => 'select',
      '#options' => image_style_options(TRUE),
      '#disabled' => !$this->moduleHandler->moduleExists('imagecache_external'),
      '#states' => array(
        'visible' => array(
          ':input[name="cache_images"]' => array('checked' => TRUE),
        ),
      ),
    ];

    $form['media']['cache_attachment_style'] = [
      '#title' => $this->t('Cache content image style'),
      '#description' => $this->t('This image style will be used for the image attachments on content in the inbox.'),
      '#default_value' => $config->get('cache_attachment_style'),
      '#type' => 'select',
      '#options' => image_style_options(TRUE),
      '#disabled' => !$this->moduleHandler->moduleExists('imagecache_external'),
      '#states' => array(
        'visible' => array(
          ':input[name="cache_images"]' => array('checked' => TRUE),
        ),
      ),
    ];

    $form['outbox'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Outbox'),
    ];

    $form['outbox']['process_outbox_handler'] = [
      '#title' => $this->t('Send activities'),
      '#type' => 'radios',
      '#options' => [
        '' => $this->t('Disabled'),
        'cron' => $this->t('On cron run'),
        'drush' => $this->t('With drush'),
      ],
      '#default_value' => $config->get('process_outbox_handler'),
      '#description' => $this->t('Activities in the Outbox are not send immediately, but are stored in a queue.<br />The drush command is <strong>activitypub:process-outbox</strong>')
    ];

    $form['inbox'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Inbox'),
    ];

    $form['inbox']['process_inbox_handler'] = [
      '#title' => $this->t('Process incoming'),
      '#type' => 'radios',
      '#options' => [
        '' => $this->t('Disabled'),
        'cron' => $this->t('On cron run'),
        'drush' => $this->t('With drush'),
      ],
      '#default_value' => $config->get('process_inbox_handler'),
      '#description' => $this->t('Activities in the Inbox that need to be processed (e.g. Create, with or without a reply) are not processed immediately, but are stored in a queue.<br />The drush command is <strong>activitypub:process-inbox</strong>')
    ];

    // Check if comment module is enabled.
    if ($this->moduleHandler->moduleExists('comment')) {
      $form['comment_create'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Create a comment'),
        '#description' => $this->t('Create a comment when an Activity arrives in the inbox of a user which has an inReplyTo property pointing to a node.<br /><br />This makes a few assumptions: the comment type is "comment", the comment body field is "comment_body", the node comment field is "comment" and there\'s an entity reference field on the comment type which maps to the Activity.<br />You can create this entity reference field using the activitypub_activity storage fields this module ships with. The other settings are usually the default when enabling the comment module.'),
      ];

      $form['comment_create']['comment_create_enable'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Enable'),
        '#default_value' => $config->get('comment_create_enable'),
      ];

      $form['comment_create']['comment_create_status'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Published'),
        '#description' => $this->t('Default status of the comment when created. To allow users to manage comments on their own content, install the <a href="https://www.drupal.org/project/manage_comment_own_content" target="_blank">Manage comments on own content</a> module.'),
        '#default_value' => $config->get('comment_create_status'),
        '#states' => [
          'visible' => ['input[name="comment_create_enable"]' => ['checked' => TRUE]],
        ],
      ];
    }

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->config('activitypub.settings')
      ->set('avatar_user_style', $form_state->getValue('avatar_user_style'))
      ->set('avatar_user_field', $form_state->getValue('avatar_user_field'))
      ->set('attachment_content_style', $form_state->getValue('attachment_content_style'))
      ->set('cache_images', $form_state->getValue('cache_images'))
      ->set('cache_avatar_style', $form_state->getValue('cache_avatar_style'))
      ->set('cache_attachment_style', $form_state->getValue('cache_attachment_style'))
      ->set('process_outbox_handler', $form_state->getValue('process_outbox_handler'))
      ->set('process_inbox_handler', $form_state->getValue('process_inbox_handler'))
      ->set('comment_create_enable', $form_state->getValue('comment_create_enable', FALSE))
      ->set('comment_create_status', $form_state->getValue('comment_create_status', TRUE))
      ->save();
    Cache::invalidateTags(['rendered']);
    parent::submitForm($form, $form_state);
  }

}
