<?php

namespace Drupal\activitypub\Plugin\activitypub\type;

use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\activitypub\Services\Type\TypePluginBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;

/**
 * The ActivityPub core types.
 *
 * @ActivityPubType(
 *   id = "activitypub_dynamic_types",
 *   label = @Translation("ActivityPub dynamic types")
 * )
 */
class DynamicTypes extends TypePluginBase {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return ['field_mapping' => []] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    $dependencies = [];

    $configuration = $this->getConfiguration();
    if (!empty($configuration['target_entity_type_id'])) {
      $dependencies['module'] = [$configuration['target_entity_type_id']];
    }

    if (!empty($configuration['target_entity_type_id']) && !empty($configuration['target_bundle'])) {
      $dependencies['config'] = [$configuration['target_entity_type_id'] . '.type.' . $configuration['target_bundle']];
    }

    return $dependencies;
  }

  /**
   * {@inheritdoc}
   */
  public function getActivities() {
    return [
      'Create', 'Like', 'Announce',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getObjects() {
    return [
      'Article', 'Note',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getProperties($object) {
    return [
      'published' => 'Date',
      'name' => 'Title',
      'content' => 'Content',
      'summary' => 'Summary',
      'object' => 'Like/Announce/Follow target',
      'inReplyTo' => 'Reply link',
      'attachment' => 'Image attachment',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $default_mapping = [];
    $default_value = $this->getConfiguration()['field_mapping'];
    foreach ($default_value as $v) {
      $default_mapping[$v['property']] = $v['field_name'];
    }

    // Get bundle fields.
    $bundle_fields = ['' => $this->t('- Select -')];
    $field_definitions = $this->entityFieldManager->getFieldDefinitions($this->getConfiguration()['target_entity_type_id'], $this->getConfiguration()['target_bundle']);
    foreach ($field_definitions as $field_definition) {
      $bundle_fields[$field_definition->getFieldStorageDefinition()
        ->getName()] = $field_definition->getName();
    }

    $form['field_mapping'] = [
      '#type' => 'container',
    ];

    $properties = $this->getProperties($this->getConfiguration()['object']);
    foreach ($properties as $property => $label) {
      $form['field_mapping'][] = [
        'property' => [
          '#type' => 'value',
          '#value' => $property,
        ],
        'field_name' => [
          '#type' => 'select',
          '#title' => $this->t('Property: @property', ['@property' => $label]),
          '#options' => $bundle_fields,
          '#default_value' => isset($default_mapping[$property]) ? $default_mapping[$property] : '',
        ],
      ];
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function build(ActivityPubActivityInterface $activity, EntityInterface $entity = NULL) {
    $object = [
      'type' => $this->getConfiguration()['object'],
      'id' => $this->renderEntityUrl($entity),
      'attributedTo' => $activity->getActor(),
    ];

    foreach ($this->getConfiguration()['field_mapping'] as $mapping) {
      if (!empty($mapping['field_name'])) {
        if ($entity->hasField($mapping['field_name']) && ($value = $entity->get($mapping['field_name'])->getValue())) {
          if ($v = $this->getValue($mapping['property'], $value)) {
            $object[$mapping['property']] = $v;
          }
        }
      }
    }

    $to = array_merge([ActivityPubActivityInterface::PUBLIC_URL], $activity->getTo());
    /** @var \Drupal\activitypub\Entity\ActivityPubActorInterface $actor */
    $actor = $this->entityTypeManager->getStorage('activitypub_actor')->loadActorByEntityIdAndType($activity->getOwnerId(), 'person');
    $cc = [$this->renderUrl(Url::fromRoute('activitypub.followers', ['user' => $actor->getOwnerId(), 'activitypub_actor' => $actor->getName()], ['absolute' => TRUE]))];
    $object['to'] = $to;
    $object['cc'] = $cc;

    // Create.
    if ($this->getConfiguration()['activity'] == 'Create') {
      $activity = [
        'type' => $this->getConfiguration()['activity'],
        'id' => $this->renderEntityUrl($activity),
        'actor' => $activity->getActor(),
        'to' => $to,
        'cc' => $cc,
        'object' => $object
      ];
    }
    else {
      $activity = [
        'type' => $this->getConfiguration()['activity'],
        'id' => $this->renderEntityUrl($activity),
        'actor' => $activity->getActor(),
        'to' => $to,
        'cc' => $cc,
        'object' => $object['object']
      ];
    }

    return $activity;
  }

}
