<?php

namespace Drupal\activitypub\Plugin\activitypub\type;

use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\activitypub\Services\Type\TypePluginBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Site\Settings;
use Drupal\Core\Url;

/**
 * The ActivityPub static types.
 *
 * @ActivityPubType(
 *   id = "activitypub_static_types",
 *   label = @Translation("ActivityPub static types")
 * )
 */
class StaticTypes extends TypePluginBase {

  /**
   * {@inheritdoc}
   */
  public function isExposed() {
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function onActivityInboxPreSave(ActivityPubActivityInterface $activity, &$doSave) {

    // Delete request. If actor and object are not the same, try to find an
    // activity with the id.
    if ($activity->getType() == 'Delete' && $activity->getCollection() == ActivityPubActivityInterface::INBOX) {
      $doSave = FALSE;

      if ($activity->getActor() != $activity->getObject()) {
        /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActivityStorageInterface $storage */
        $storage = $this->entityTypeManager->getStorage('activitypub_activity');
        foreach ($storage->getActivities(['object' => $activity->getObject(), '!type' => 'Delete']) as $a) {
          $a->delete();
        }
      }
    }

    // Check if the actor already exists. If so, do not save the activity,
    // otherwise, set status to 0.
    if ($activity->getType() == 'Follow') {
      /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActivityStorageInterface $storage */
      $storage = $this->entityTypeManager->getStorage('activitypub_activity');
      $activities = $storage->getActivities(['type' => 'Follow', 'actor' => $activity->getActor(), 'object' => $activity->getObject(), 'collection' => ActivityPubActivityInterface::INBOX]);
      if (!empty($activities)) {
        $doSave = FALSE;
      }
      else {
        $activity->set('status', FALSE);
      }
    }

    // Check undo request.
    if ($activity->getType() == 'Undo') {
      $payload = @json_decode($activity->getPayLoad(), TRUE);
      if (isset($payload['object']) && isset($payload['object']['type']) && $payload['object']['type'] == 'Follow') {
        $doSave = FALSE;
        /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActivityStorageInterface $storage */
        $storage = $this->entityTypeManager->getStorage('activitypub_activity');
        foreach ($storage->getActivities(['type' => 'Follow', 'actor' => $activity->getActor(), 'object' => $activity->getObject()]) as $a) {
          $a->delete();
        }
        foreach ($storage->getActivities(['type' => 'Accept', 'object' => $activity->getActor(), 'actor' => $activity->getObject()]) as $a) {
          $a->delete();
        }
      }
    }

  }

  /**
   * {@inheritdoc}
   */
  public function onActivityPostSave(ActivityPubActivityInterface $activity, $update = TRUE) {

    // Incoming follow request.
    if ($activity->getType() == 'Follow' && $activity->getCollection() == ActivityPubActivityInterface::INBOX && !$update) {

      $values = [
        'uid' => $activity->getOwnerId(),
        'collection' => ActivityPubActivityInterface::OUTBOX,
        'config_id' => 'accept',
        'type' => 'Accept',
        'actor' => $activity->getObject(),
        'object' => $activity->getActor(),
        'external_id' => $activity->getExternalId(),
        'status' => 0,
      ];

      /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $outboxActivity */
      $outboxActivity = $this->entityTypeManager->getStorage('activitypub_activity')->create($values);
      $outboxActivity->save();
      $this->activityPubProcessClient->createQueueItem($outboxActivity);
    }

    // Outgoing follow request.
    if ($activity->getType() == 'Follow' && $activity->getCollection() == ActivityPubActivityInterface::OUTBOX && !$update) {
      $this->activityPubProcessClient->createQueueItem($activity);
    }

    // Outgoing accept request.
    if ($activity->getType() == 'Accept' && $activity->getCollection() == ActivityPubActivityInterface::OUTBOX && $update && $activity->isProcessed()) {
      /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActivityStorageInterface $storage */
      $storage = $this->entityTypeManager->getStorage('activitypub_activity');
      $activities = $storage->getActivities(['type' => 'Follow', 'actor' => $activity->getObject(), 'object' => $activity->getActor(), 'collection' => ActivityPubActivityInterface::INBOX]);
      if (!empty($activities)) {
        $followActivity = array_shift($activities);
        $followActivity->set('status', TRUE);
        $followActivity->save();
      }
    }

    // Incoming accept request.
    if ($activity->getType() == 'Accept' && $activity->getCollection() == ActivityPubActivityInterface::INBOX && !$update) {
      /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActivityStorageInterface $storage */
      $storage = $this->entityTypeManager->getStorage('activitypub_activity');
      $activities = $storage->getActivities(['type' => 'Follow', 'actor' => $activity->getObject(), 'object' => $activity->getActor(), 'collection' => ActivityPubActivityInterface::OUTBOX]);
      if (!empty($activities)) {
        $followActivity = array_shift($activities);
        $followActivity->set('status', TRUE);
        $followActivity->save();
      }
    }

    // Undo request.
    if ($activity->getType() == 'Undo' && !$update) {
      $payload = @json_decode($activity->getPayLoad(), TRUE);
      if (isset($payload['object']) && isset($payload['object']['type']) && $payload['object']['type'] == 'Follow') {
        foreach ($this->entityTypeManager->getStorage('activitypub_activity')->loadByProperties(['type' => 'Follow', 'actor' => $activity->getActor(), 'object' => $activity->getObject()]) as $a) {
          $a->delete();
        }
      }
    }

    // Create comment.
    if ($activity->getType() == 'Create' && $this->configFactory->get('activitypub.settings')->get('comment_create_enable') && !$update) {
      $payload = @json_decode($activity->getPayLoad(), TRUE);
      if (isset($payload['object']) && isset($payload['object']['inReplyTo']) && !empty($payload['object']['inReplyTo'])) {
        $activity->set('config_id', 'inbox_reply');
        $this->activityPubProcessClient->createQueueItem($activity);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function build(ActivityPubActivityInterface $activity, EntityInterface $entity = NULL) {
    $return = [];

    // Follow type.
    if ($this->getConfiguration()['activity'] == 'Follow') {
      $attributes = [
        'id' => $this->renderEntityUrl($activity),
        'type' => $this->getConfiguration()['activity'],
        'actor' => $activity->getActor(),
        'object' => $activity->getObject(),
        'to' => [$activity->getObject()],
      ];
      $return = $attributes;
    }

    // Accept type.
    if ($this->getConfiguration()['activity'] == 'Accept') {
      $attributes = [
        'id' => $this->renderEntityUrl($activity),
        'type' => $this->getConfiguration()['activity'],
        'actor' => $activity->getActor(),
        'to' => [$activity->getObject()],
        'object' => [
          'type' => 'Follow',
          'id' => $activity->getExternalId(),
          'actor' => $activity->getObject(),
          'object' => $activity->getActor(),
        ]
      ];
      $return = $attributes;
    }

    // Undo type.
    if ($this->getConfiguration()['activity'] == 'Undo') {
      $attributes = [
        'id' => $this->renderEntityUrl($activity),
        'type' => $this->getConfiguration()['activity'],
        'actor' => $activity->getActor(),
        'to' => [$activity->getObject()],
        'object' => [
          'type' => 'Follow',
          'id' => $activity->getExternalId(),
          'actor' => $activity->getActor(),
          'object' => $activity->getObject(),
        ]
      ];
      $return = $attributes;
    }

    // Delete type.
    if ($this->getConfiguration()['activity'] == 'Delete') {

      $to = array_merge([ActivityPubActivityInterface::PUBLIC_URL], $activity->getTo());
      /** @var \Drupal\activitypub\Entity\ActivityPubActorInterface $actor */
      $actor = $this->entityTypeManager->getStorage('activitypub_actor')->loadActorByEntityIdAndType($activity->getOwnerId(), 'person');
      $cc = [$this->renderUrl(Url::fromRoute('activitypub.followers', ['user' => $actor->getOwnerId(), 'activitypub_actor' => $actor->getName()], ['absolute' => TRUE]))];
      $attributes['to'] = $to;
      $attributes['cc'] = $cc;

      $id = $this->renderEntityUrl($activity);
      $object_id = $activity->getObject();

      $attributes = [
        'type' => $this->getConfiguration()['activity'],
        'actor' => $activity->getActor(),
        'to' => $to,
        'cc' => $cc,
        'id' => $id,
        'object' => ['id' => $object_id],
      ];
      $return = $attributes;
    }

    return $return;
  }

  /**
   * {@inheritdoc}
   */
  public function doInboxProcess(ActivityPubActivityInterface $activity, EntityInterface $entity = NULL) {
    if ($activity->getType() == 'Create' && $this->configFactory->get('activitypub.settings')->get('comment_create_enable')) {
      $payload = @json_decode($activity->getPayLoad(), TRUE);
      if (isset($payload['object']) && isset($payload['object']['inReplyTo']) && !empty($payload['object']['inReplyTo'])) {
        /** @noinspection PhpUnhandledExceptionInspection */
        // The exception is caught in the process client.
        $this->handleCommentCreation($activity);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function onEntityDelete(ActivityPubActivityInterface $activity, EntityInterface $entity) {

    $values = [
      'uid' => $activity->getOwnerId(),
      'collection' => ActivityPubActivityInterface::OUTBOX,
      'config_id' => 'delete',
      'type' => 'Delete',
      'actor' => $activity->getActor(),
      'object' => $this->renderEntityUrl($entity),
      'external_id' => $this->renderEntityUrl($activity),
      'status' => 0,
      'to' => $activity->getTo(),
    ];

    /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $outboxActivity */
    $outboxActivity = $this->entityTypeManager->getStorage('activitypub_activity')->create($values);
    $outboxActivity->save();
    $this->activityPubProcessClient->createQueueItem($outboxActivity);
  }

  /**
   * Handle comment creation.
   *
   * @param \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity
   */
  protected function handleCommentCreation(ActivityPubActivityInterface $activity) {
    $pid = 0;

    $replyPath = trim(parse_url($activity->getObject(), PHP_URL_PATH));
    if (empty($replyPath)) {
      $replyPath = "/";
    }

    try {
      $path = \Drupal::service('path_alias.manager')->getPathByAlias($replyPath);
      $params = Url::fromUri("internal:" . $path)->getRouteParameters();

      // This can be a reply on a comment. Get the node to attach the
      // comment there and set pid.
      if (!empty($params) && key($params) == 'comment') {
        /** @var \Drupal\comment\CommentInterface $comment */
        $comment = $this->entityTypeManager->getStorage('comment')->load($params['comment']);
        if ($comment && $comment->getCommentedEntityTypeId() == 'node') {
          $pid = $comment->id();
          $params = ['node' => $comment->getCommentedEntityId()];
        }
      }

      // We currently support only comments on nodes.
      if (!empty($params) && key($params) == 'node') {
        $entity_type = key($params);

        $storage = $this->entityTypeManager->getStorage($entity_type);
        if ($storage) {

          $payload = @json_decode($activity->getPayLoad(), TRUE);
          $object = $payload['object'];

          $comment_type = 'comment';
          $activity_reference_field = 'activitypub_activity';
          $node_comment_field_name = 'comment';

          /** @var \Drupal\node\NodeInterface $node */
          $node = $storage->load($params[$entity_type]);
          if ($node && $node->hasField($node_comment_field_name) && $node->{$node_comment_field_name}->status == 2) {

            $subject = 'Reply by ' . $payload['actor'];

            $values = [
              'subject' => $subject,
              'name' => $payload['actor'],
              'status' => $this->configFactory->get('activitypub.settings')->get('comment_create_status'),
              'entity_id' => $node->id(),
              'entity_type' => 'node',
              'pid' => $pid,
              'comment_body' => [
                'value' => $object['content'],
                'format' => Settings::get('activitypub_comment_format', 'restricted_html'),
              ],
              'comment_type' => $comment_type,
              'field_name' => $node_comment_field_name,
              $activity_reference_field => [
                'target_id' => $activity->id(),
              ],
            ];

            $comment = $this->entityTypeManager->getStorage('comment')->create($values);
            $comment->setCreatedTime($activity->getCreatedTime());
            $comment->save();
          }
        }
      }
    }
    catch (\Exception $e) {
      $this->logger->notice('Error creating comment: @message', ['@message' => $e->getMessage()]);
    }
  }

}
