<?php

namespace Drupal\activitypub\Services;

use Drupal\Core\File\FileSystemInterface;
use phpseclib\Crypt\RSA;
use Psr\Log\LoggerInterface;

class ActivityPubKeys implements ActivityPubKeysInterface {

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * ActivityPubKeys constructor
   *
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger
   */
  public function __construct(FileSystemInterface $file_system, LoggerInterface $logger) {
    $this->fileSystem = $file_system;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public function getPublicKey($path) {
    $key_path = rtrim(activitypub_keys_path(), '/') . '/' . $path . '/public.pem';
    return file_get_contents($key_path);
  }

  /**
   * {@inheritdoc}
   */
  public function getPrivateKey($path) {
    $key_path = rtrim(activitypub_keys_path(), '/') . '/' . $path . '/private.pem';
    return file_get_contents($key_path);
  }

  /**
   * {@inheritdoc}
   */
  public function createSignature($private_key_path, $host, $path, $digest, $date = NULL) {
    if (!isset($date)) {
      $date = gmdate('D, d M Y H:i:s T', time());
    }

    $rsa = new RSA();
    $rsa->loadKey($this->getPrivateKey($private_key_path));
    $plaintext = "(request-target): post $path\nhost: $host\ndate: $date\ndigest: $digest";
    $rsa->setHash("sha256");
    $rsa->setSignatureMode(RSA::SIGNATURE_PKCS1);
    return $rsa->sign($plaintext);
  }

  /**
   * {@inheritdoc}
   */
  public function createDigest($message) {
    return 'SHA-256=' . base64_encode(hash('sha256', $message, true));
  }

  /**
   * {@inheritdoc}
   */
  public function generateKeys($path) {
    $return = FALSE;

    try {

      $rsa = new RSA();
      $keys = $rsa->createKey(4096);

      $dir_path = rtrim(activitypub_keys_path(), '/') . '/' . $path;
      $this->fileSystem->prepareDirectory($dir_path, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);

      foreach (['public' => 'publickey', 'private' => 'privatekey'] as $filename => $rsakey) {

        // Key uri.
        $key_uri = "$dir_path/$filename.pem";

        // Write key content to key file.
        $this->fileSystem->saveData($keys[$rsakey], $key_uri, FileSystemInterface::EXISTS_REPLACE);

        // Set correct permission to key file.
        $this->fileSystem->chmod($key_uri, 0600);

      }

      $return = TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error('Error generating keys: @message', ['@message' => $e->getMessage()]);
    }

    return $return;

  }

  /**
   * {@inheritdoc}
   */
  public function deleteKeys($path) {
    $dir_path = rtrim(activitypub_keys_path(), '/') . '/' . $path;
    $this->fileSystem->deleteRecursive($dir_path);
  }

}
