<?php

namespace Drupal\activitypub\Services;

use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use Psr\Log\LoggerInterface;

class ActivityPubProcessClient implements ActivityPubProcessClientInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The ActivityPub Utility service.
   *
   * @var \Drupal\activitypub\Services\ActivityPubUtilityInterface
   */
  protected $activityPubUtility;

  /**
   * The ActivityPub keys service.
   *
   * @var \Drupal\activitypub\Services\ActivityPubKeysInterface
   */
  protected $activityPubKeys;

  /**
   * The http client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Whether in debug mode or not.
   *
   * @var bool
   */
  protected $debug = FALSE;

  /**
   * ActivityPubProcessClient constructor
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\activitypub\Services\ActivityPubUtilityInterface $activitypub_utility
   *   The activitypub utility.
   * @param \Drupal\activitypub\Services\ActivityPubKeysInterface $activitypub_keys
   *   The activitypub keys.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The HTTP client.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, ActivityPubUtilityInterface $activitypub_utility, ActivityPubKeysInterface $activitypub_keys, ClientInterface $http_client, LoggerInterface $logger) {
    $this->entityTypeManager = $entity_type_manager;
    $this->activityPubUtility = $activitypub_utility;
    $this->activityPubKeys = $activitypub_keys;
    $this->httpClient = $http_client;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public function handleOutboxQueue($send = TRUE, $remove_item_from_queue = TRUE, $debug = FALSE) {
    $end = time() + 15;
    $release_items = [];
    $this->setDebug($debug);
    $server = $this->activityPubUtility->getServer();

    /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorage $actorStorage */
    $actorStorage = $this->entityTypeManager->getStorage('activitypub_actor');

    while (time() < $end && ($item = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->claimItem())) {
      $data = $item->data;

      if (empty($data['activity'])) {
        $this->deleteItemFromQueue($item, ACTIVITYPUB_OUTBOX_QUEUE);
        continue;
      }

      $remove_queue_item = TRUE;
      if (!$remove_item_from_queue) {
        $remove_queue_item = FALSE;
      }

      $activity = NULL;
      try {
        /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity */
        $activity = $this->entityTypeManager->getStorage('activitypub_activity')->load($data['activity']);

        // Check the queue value.
        if (!$activity->isQueued()) {
          $this->deleteItemFromQueue($item, ACTIVITYPUB_OUTBOX_QUEUE);
          continue;
        }

        $actor = $actorStorage->loadActorByEntityIdAndType($activity->getOwnerId(), 'person');

        // Build activity.
        $build = $activity->buildActivity();

        // Add context.
        $build['@context'] = ActivityPubActivityInterface::CONTEXT_URL;

        // Send to.
        $inboxes = [];
        $targets = [];
        if (!empty($build['to'])) {
          foreach ($build['to'] as $t) {
            if ($t != ActivityPubActivityInterface::PUBLIC_URL) {
              $targets[] = $t;
            }
          }
        }

        // cc, these are always followers at the moment.
        if (!empty($build['cc'][0])) {

          if (isset($build['object']['cc'])) {
            $build['object']['cc'] = $build['cc'];
          }

          // Get followers.
          $conditions = ['type' => 'Follow', 'object' => str_replace('/followers', '', $build['cc'][0]), 'status' => 1];
          /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $followers */
          $followers = $this->entityTypeManager->getStorage('activitypub_activity')->loadByProperties($conditions);
          foreach ($followers as $follower) {
            $targets[] = $follower->getActor();
          }
        }

        $this->debug($build);

        // Get inboxes.
        foreach ($targets as $target) {
          $target_actor = $server->actor($target);
          if ($target_actor) {
            $inbox = (string) $target_actor->get('inbox');
            if (($endpoints = $target_actor->get('endpoints')) && !empty($endpoints['sharedInbox'])) {
              $inbox = $endpoints['sharedInbox'];
            }
            $inboxes[$inbox] = $inbox;
          }
        }

        if (!empty($inboxes)) {
          $this->debug($inboxes);
          $keyId = $activity->getActor();

          // Create digest.
          $json = json_encode($build);
          $digest = $this->activityPubKeys->createdigest($json);

          foreach ($inboxes as $inbox) {

            $parsed = parse_url($inbox);
            $host = $parsed['host'];
            $path = $parsed['path'];
            $date = gmdate('D, d M Y H:i:s T', time());
            $this->debug($parsed);

            // Create signature.
            $signature = $this->activityPubKeys->createSignature($actor->getName(), $host, $path, $digest, $date);

            $headers = [
              'Content-Type' => 'application/activity+json',
              'host' => $host,
              'date' => $date,
              'digest' => $digest,
              'Signature' => 'keyId="' . $keyId . '#main-key",headers="(request-target) host date digest",signature="' . base64_encode($signature) . '"',
            ];

            $this->debug($headers);

            if ($send) {
              try {
                $response = $this->httpClient->post($inbox, ['body' => $json, 'headers' => $headers]);
                $this->logger->notice('Outbox response to @inbox for @id: code: @code -  @response', ['@inbox' => $inbox, '@id' => $activity->id(), '@code' => $response->getStatusCode(), '@response' => (string) $response->getBody()]);
              }
              catch (RequestException $e) {
                $this->logger->error('Outbox exception to @inbox for @id to @inbox: @message', ['@inbox' => $inbox, '@message' => $e->getMessage(), '@id' => $activity->id()]);
              }
            }
          }
        }
      }
      catch (\Exception $e) {
        $this->debug($e->getMessage());
        $this->logger->error('Outbox general exception to @inbox for @id: @message', ['@message' => $e->getMessage(), '@id' => $activity->id()]);
      }

      // Remove or release the queue item.
      if ($remove_queue_item) {
        if ($activity) {
          try {
            $activity->set('processed', TRUE);
            $activity->set('queued', FALSE);
            $activity->save();
          }
          catch (\Exception $ignored) {}
        }
        $this->deleteItemFromQueue($item, ACTIVITYPUB_OUTBOX_QUEUE);
      }
      else {
        $release_items[] = $item;
      }
    }

    foreach ($release_items as $item) {
      $this->releaseItemFromQueue($item, ACTIVITYPUB_OUTBOX_QUEUE);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function handleInboxQueue($remove_item_from_queue = TRUE, $debug = FALSE) {
    $end = time() + 15;
    $release_items = [];
    $this->setDebug($debug);

    while (time() < $end && ($item = \Drupal::queue(ACTIVITYPUB_INBOX_QUEUE)->claimItem())) {
      $data = $item->data;

      if (empty($data['activity'])) {
        $this->deleteItemFromQueue($item, ACTIVITYPUB_INBOX_QUEUE);
        continue;
      }

      $activity = NULL;
      $remove_queue_item = TRUE;

      try {
        /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity */
        $activity = $this->entityTypeManager->getStorage('activitypub_activity')->load($data['activity']);

        // Check the queue value.
        if (!$activity->isQueued()) {
          $this->deleteItemFromQueue($item, ACTIVITYPUB_OUTBOX_QUEUE);
          continue;
        }

        $activity->doInboxProcess();

        if (!$remove_item_from_queue) {
          $remove_queue_item = FALSE;
        }

      }
      catch (\Exception $e) {
        $this->debug($e->getMessage());
        $this->logger->error('Inbox general exception for @id: @message', ['@message' => $e->getMessage(), '@id' => $activity->id()]);
      }

      // Remove or release the queue item.
      if ($remove_queue_item) {
        if ($activity) {
          try {
            $activity->set('processed', TRUE);
            $activity->set('queued', FALSE);
            $activity->save();
          }
          catch (\Exception $ignored) {}
        }
        $this->deleteItemFromQueue($item, ACTIVITYPUB_INBOX_QUEUE);
      }
      else {
        $release_items[] = $item;
      }

    }

    foreach ($release_items as $item) {
      $this->releaseItemFromQueue($item, ACTIVITYPUB_INBOX_QUEUE);
    }

  }

  /**
   * {@inheritdoc}
   */
  public function createQueueItem(ActivityPubActivityInterface $activityPubActivity) {

    $data = [
      'activity' => $activityPubActivity->id(),
    ];

    try {
      $queue = $activityPubActivity->getCollection() == 'outbox' ? ACTIVITYPUB_OUTBOX_QUEUE : ACTIVITYPUB_INBOX_QUEUE;
      \Drupal::queue($queue)->createItem($data);
      $activityPubActivity->set('queued', TRUE)->save();
    }
    catch (\Exception $e) {
      $this->logger->notice('Error creating queue item: @message', ['@message' => $e->getMessage()]);
    }
  }

  /**
   * Delete item from queue.
   *
   * @param $item
   * @param $queue
   */
  protected function deleteItemFromQueue($item, $queue) {
    \Drupal::queue($queue)->deleteItem($item);
  }

  /**
   * Release item from queue.
   *
   * @param $item
   * @param $queue
   */
  protected function releaseItemFromQueue($item, $queue) {
    \Drupal::queue($queue)->releaseItem($item);
  }

  /**
   * Debug helper function.
   *
   * @param $var
   */
  protected function debug($var) {
    if ($this->debug) {
      print_r($var);
      echo "\n--------\n";
    }
  }

  /**
   * Set debug property.
   *
   * @param $debug
   */
  protected function setDebug($debug) {
    $this->debug = $debug;
  }

}
