<?php

namespace Drupal\activitypub\Services;

use ActivityPhp\Server;
use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Pager\PagerManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\reader\ReaderBase;
use Symfony\Component\HttpFoundation\RequestStack;

class Reader extends ReaderBase {

  use StringTranslationTrait;

  /**
   * The ActivityPub Actor storage.
   *
   * @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorageInterface
   */
  protected $actorStorage;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The pager manager.
   *
   * @var \Drupal\Core\Pager\PagerManagerInterface
   */
  protected $pagerManager;

  /**
   * The request stack
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack $requestStack
   */
  protected $requestStack;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\Datetime\DateFormatInterface
   */
  protected $dateFormatter;

  /**
   * The ActivityPub Utility service.
   *
   * @var \Drupal\activitypub\Services\ActivityPubUtilityInterface
   */
  protected $activityPubUtility;

  /**
   * The ActivityPub Media cache service.
   *
   * @var \Drupal\activitypub\Services\ActivityPubMediaCacheInterface
   */
  protected $activityPubMediaCache;

  /**
   * ActivityPubFormAlter constructor
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   * @param \Drupal\Core\Pager\PagerManagerInterface $pager_manager
   *   The pager manager.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   The date formatter.
   * @param \Drupal\activitypub\Services\ActivityPubUtilityInterface $activitypub_utility
   *   The ActivityPub utility service.
   * @param \Drupal\activitypub\Services\ActivityPubMediaCacheInterface $activitypub_media_cache
   *   The ActivityPub media cache service.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, AccountInterface $current_user, RequestStack $request_stack, PagerManagerInterface $pager_manager, DateFormatterInterface $date_formatter, ActivityPubUtilityInterface $activitypub_utility, ActivityPubMediaCacheInterface $activitypub_media_cache) {
    $this->entityTypeManager = $entity_type_manager;
    $this->actorStorage = $entity_type_manager->getStorage('activitypub_actor');
    $this->currentUser = $current_user;
    $this->requestStack = $request_stack;
    $this->pagerManager = $pager_manager;
    $this->dateFormatter = $date_formatter;
    $this->activityPubUtility = $activitypub_utility;
    $this->activityPubMediaCache = $activitypub_media_cache;
  }

  /**
   * {@inheritdoc}
   */
  public function getChannels() {

    if (!$this->getActor()) {
      return [];
    }

    $conditions = [
      'uid' => $this->currentUser->id(),
      'status' => 1,
      'collection' => 'inbox',
      'is_read' => 0,
      'type' => ['Create', 'Like', 'Announce'],
    ];
    $unread = $this->entityTypeManager->getStorage('activitypub_activity')->getActivityCount($conditions);

    return [
      'label' => $this->t('Fediverse'),
      'channels' => [
        (object) [
          'uid' => 'local',
          'unread' => $unread,
          'name' => $this->t('Local timeline')
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getTimelineActions($id) {

    if (!$this->getActor()) {
      return [];
    }

    return [[
      'action' => 'mark-read',
      'title' => $this->t('Mark read'),
    ]];
  }

  /**
   * {@inheritdoc}
   */
  public function doTimelineAction($action, $id) {
    if ($this->getActor()) {
      if ($action == 'mark-read') {
        $this->entityTypeManager->getStorage('activitypub_activity')->changeReadStatus(1);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getTimeline($id, $search = NULL) {

    if (!$this->getActor()) {
      return ['items' => []];
    }

    return $this->getResponse($id, $search);
  }

  /**
   * Get the actor for the current user.
   *
   * @return bool
   */
  protected function getActor() {
    return $this->currentUser->hasPermission('access reader') && (bool) $this->actorStorage->loadActorByEntityIdAndType($this->currentUser->id(), 'person');
  }

  /**
   * Get items for a timeline.
   *
   * @param $id
   *   The channel id
   * @param $search
   *   The search param.
   *
   * @return array
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getResponse($id, $search) {

    $items = [];
    $paging = [];

    // Get page.
    $page = $this->requestStack->getCurrentRequest()->get('page', 0);
    $server = $this->activityPubUtility->getServer();

    /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface[] $activities */
    $storage = $this->entityTypeManager->getStorage('activitypub_activity');
    $ids = $storage->getQuery()
      ->condition('uid', $this->currentUser->id())
      ->condition('status', 1)
      ->condition('collection', 'inbox')
      ->condition('type', $this->activityPubUtility->getTimelineTypes(), 'IN')
      ->pager(20)
      ->sort('id', 'DESC')
      ->execute();
    $activities = $storage->loadMultiple($ids);
    foreach ($activities as $activity) {
      if ($item = $this->buildMicrosubItem($activity, $server)) {
        $items[] = $item;
      }
    }

    $pager = $this->pagerManager->getPager();
    $pager_total = $pager->getTotalPages();
    $page++;
    if ((isset($pager_total) && is_numeric($pager_total) && $pager_total > $page) || $page > 0) {
      $paging = ['after' => $page];
    }

    $response = ['items' => $items];
    if (!empty($paging)) {
      $response['paging'] = (object) $paging;
    }

    return $response;
  }

  /**
   * Build a Microsub item from an activity payload.
   *
   * @param \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity
   * @param \ActivityPhp\Server $server
   *
   * @return \stdClass $item
   */
  protected function buildMicrosubItem(ActivityPubActivityInterface $activity, Server $server) {

    $payload = json_decode($activity->getPayLoad(), TRUE);
    $context = json_decode($activity->getContext(), TRUE);

    $item = [];
    $item['_id'] = $activity->id();
    $item['_is_read'] = $activity->isRead();
    $item['type'] = 'entry';
    $item['url'] = $activity->getObject();
    $item['published'] = $this->dateFormatter->format($activity->getCreatedTime(), 'custom','Y-m-dTH:i:s');

    // Content and response type.
    if ($activity->getType() == 'Like') {
      $item['like-of'] = [$activity->getObject()];
    }
    elseif ($activity->getType() == 'Announce') {
      $item['repost-of'] = [$activity->getObject()];
    }
    elseif ($activity->getType() == 'Create') {
      if (is_array($payload['object'])) {
        if (!empty($payload['object']['inReplyTo'])) {
          $item['in-reply-to'] = [$payload['object']['inReplyTo']];
        }

        if (!empty($payload['object']['name'])) {
          $item['name'] = $payload['object']['name'];
        }

        if (!empty($payload['object']['content'])) {
          $item['content'] = (object) ['html' => $payload['object']['content']];
        }

        if (!empty($payload['object']['attachment'])) {
          $this->handleAttachments($payload['object']['attachment'], $item);
        }

      }
      else {
        if (!empty($payload['name'])) {
          $item['name'] = $payload['name'];
        }
        else {
          // currently treat as repost
          $item['repost-of'] = [$activity->getObject()];
        }
      }
    }

    // Context.
    if (!empty($context) && (isset($item['repost-of']) || isset($item['like-of']) || isset($item['in-reply-to']))) {
      $url = NULL;
      if (isset($item['repost-of'])) {
        $url = $item['repost-of'][0];
      }
      elseif (isset($item['like-of'])) {
        $url = $item['like-of'][0];
      }
      elseif (isset($item['in-reply-to'])) {
        $url = $item['in-reply-to'][0];
      }

      if (isset($url) && isset($context['id']) && $context['id'] == $url) {
        $ref = [];
        if (!empty($context['name'])) {
          $ref['name'] = $context['name'];
        }
        if (!empty($context['content'])) {
          $ref['content'] = (object) ['html' => $context['content']];
        }
        if (!empty($context['summary'])) {
          $ref['summary'] = $context['summary'];
        }
        if (!empty($context['attachment'])) {
          $this->handleAttachments($context['attachment'], $ref);
        }

        $item['refs'] = new \stdClass();
        $item['refs']->{$url} = (object) $ref;
      }
    }

    // Author information.
    $name = $photo = '';
    $target_actor = $server->actor($activity->getActor());
    if ($target_actor) {
      $name = $target_actor->get('name');
      $icon = $target_actor->get('icon');
      if (!empty($icon)) {
        $photo = $this->activityPubMediaCache->applyImageCache($icon->get('url'));
      }
    }

    $item['author'] = (object) ['url' => $activity->getActor(), 'name' => $name, 'photo' => $photo];

    return (object) $item;
  }

  /**
   * Handle attachments.
   *
   * @param $attachments
   * @param $o
   */
  protected function handleAttachments($attachments, &$o) {
    foreach ($attachments as $attachment) {
      if (isset($attachment['mediaType'])) {
        switch ($attachment['mediaType']) {
          case 'image/jpg':
          case 'image/png':
          case 'image/jpeg':
          case 'image/gif':
            if (!isset($o['photo'])) {
              $o['photo'] = [];
            }
            $o['photo'][] = $this->activityPubMediaCache->applyImageCache($attachment['url'], 'cache_attachment');
            break;
          case 'video/mp4':
            if (!isset($o['video'])) {
              $o['video'] = [];
            }
            $o['video'][] = $attachment['url'];
            break;
          case 'audio/mp3':
            if (!isset($o['audio'])) {
              $o['audio'] = [];
            }
            $o['audio'][] = $attachment['url'];
            break;
        }
      }
    }
  }

}
