<?php

namespace Drupal\activitypub\Services\Type;

use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\activitypub\Services\ActivityPubProcessClientInterface;
use Drupal\activitypub\Services\ActivityPubUtilityInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\Render\RenderContext;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Url;
use Drupal\file\FileInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A base class for ActivityPub types that implements most of the boilerplate.
 */
abstract class TypePluginBase extends PluginBase implements TypePluginInterface, ContainerFactoryPluginInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The renderer.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * The date formatter.
   *
   * @var \Drupal\Core\DateTime\DateFormatInterface
   */
  protected $dateFormatter;

  /**
   * The ActivityPub process client.
   *
   * @var \Drupal\activitypub\Services\ActivityPubProcessClient
   */
  protected $activityPubProcessClient;

  /**
   * The ActivityPub utility service.
   *
   * @var \Drupal\activitypub\Services\ActivityPubUtilityInterface
   */
  protected $activityPubUtility;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, RendererInterface $renderer, DateFormatterInterface $date_formatter, ActivityPubProcessClientInterface $activitypub_process_client, ActivityPubUtilityInterface $activitypub_utility, EntityFieldManagerInterface $entity_field_manager, ConfigFactoryInterface $config_factory, LoggerInterface $logger) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->setConfiguration($configuration);
    $this->entityTypeManager = $entity_type_manager;
    $this->renderer = $renderer;
    $this->dateFormatter = $date_formatter;
    $this->activityPubProcessClient = $activitypub_process_client;
    $this->activityPubUtility = $activitypub_utility;
    $this->entityFieldManager = $entity_field_manager;
    $this->configFactory = $config_factory;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('renderer'),
      $container->get('date.formatter'),
      $container->get('activitypub.process.client'),
      $container->get('activitypub.utility'),
      $container->get('entity_field.manager'),
      $container->get('config.factory'),
      $container->get('logger.channel.activitypub')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function isExposed() {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function build(ActivityPubActivityInterface $activity, EntityInterface $entity = NULL) {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getActivities() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getObjects() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getProperties($object) {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function onActivityInboxPreSave(ActivityPubActivityInterface $activity, &$doSave) {}

  /**
   * {@inheritdoc}
   */
  public function onActivityPostSave(ActivityPubActivityInterface $activity, $update = TRUE) {}

  /**
   * {@inheritdoc}
   */
  public function doInboxProcess(ActivityPubActivityInterface $activity, EntityInterface $entity = NULL) {}

  /**
   * {@inheritdoc}
   */
  public function onEntityDelete(ActivityPubActivityInterface $activity, EntityInterface $entity) {}

  /**
   * Get value for a property.
   *
   * @param $property
   * @param $value
   *
   * @return array|mixed|null
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getValue($property, $value) {
    $return = NULL;

    switch ($property) {
      case 'published':
        $return = $this->dateFormatter->format($value[0]['value'], 'custom', 'Y-m-d\TH:i:s\Z');
        break;
      case 'name':
      case 'content':
      case 'summary':
        $return = trim($value[0]['value']);
        break;
      case 'object':
      case 'inReplyTo':
        $return = $value[0]['uri'];
        break;
      case 'attachment':
        $return = [];
        foreach ($value as $v) {
          /** @var \Drupal\file\FileInterface $file */
          $file = $this->entityTypeManager->getStorage('file')->load($v['target_id']);
          $return[] = (object) [
            'type' => 'Image',
            'mediaType' => $file->getMimeType(),
            'url' => $this->getFileUrl($file),
          ];
        }
        break;
    }

    return $return;
  }

  /**
   * Gets the file URL.
   *
   * @param \Drupal\file\FileInterface $file
   *
   * @return string
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getFileUrl(FileInterface $file) {
    $config = $this->configFactory->get('activitypub.settings');
    $attachment_content_style = $config->get('attachment_content_style');
    if (!empty($attachment_content_style)) {
      /** @var \Drupal\image\ImageStyleInterface $image_style */
      $storage = $this->entityTypeManager->getStorage('image_style');
      $image_style = $storage->load($attachment_content_style);
      return $image_style->buildUrl($file->getFileUri());
    }
    else {
      return file_create_url($file->getFileUri());
    }
  }

  /**
   * Render a URL from an entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *
   * @return string
   */
  protected function renderEntityUrl(EntityInterface $entity) {
    return $this->renderer->executeInRenderContext(new RenderContext(), function () use ($entity) {
      return $entity->toUrl('canonical', ['absolute' => TRUE])->toString();
    });
  }

  /**
   * Render a Url.
   *
   * @param \Drupal\Core\Url $url
   *
   * @return string
   */
  protected function renderUrl(Url $url) {
    return $this->renderer->executeInRenderContext(new RenderContext(), function () use ($url) {
      return $url->toString();
    });
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'target_entity_type_id' => 'node',
      'target_bundle' => '',
      'activity' => '',
      'object' => '',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = NestedArray::mergeDeep(
      $this->defaultConfiguration(),
      $configuration
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $entity_type_id = $this->getConfiguration()['target_entity_type_id'];

    $entity_types = ['node' => $this->t('Content')];
    if (\Drupal::moduleHandler()->moduleExists('comment')) {
      $entity_types['comment'] = $this->t('Comment');
    }
    $form['target_entity_type_id'] = [
      '#type' => 'select',
      '#title' => $this->t('Entity type'),
      '#options' => $entity_types,
      '#default_value' => $entity_type_id,
      '#ajax' => [
        'trigger_as' => ['name' => 'type_configure'],
        'callback' => '::buildAjaxTypeForm',
        'wrapper' => 'type-config-form',
        'method' => 'replace',
        'effect' => 'fade',
      ],
    ];

    if ($entity_type_id == 'comment') {
      $bundle_options = activitypub_comment_get_names();
    }
    else {
      $bundle_options = node_type_get_names();
    }
    $form['target_bundle'] = [
      '#type' => 'select',
      '#title' => $this->t('Type'),
      '#options' => ['' => $this->t('- Select -')] + $bundle_options,
      '#default_value' => $this->getConfiguration()['target_bundle'],
      '#ajax' => [
        'trigger_as' => ['name' => 'type_configure'],
        'callback' => '::buildAjaxTypeForm',
        'wrapper' => 'type-config-form',
        'method' => 'replace',
        'effect' => 'fade',
      ],
    ];

    $activitypub_activity_options = ['' => $this->t('- Select -')];
    $activitypub_object_options = ['' => $this->t('- Select -')];

    foreach ($this->getActivities() as $a) {
      $activitypub_activity_options[$a] = $a;
    }
    foreach ($this->getObjects() as $o) {
      $option_id = $o;
      $activitypub_object_options[$option_id] = $o;
    }

    $form['activity'] = [
      '#type' => 'select',
      '#title' => $this->t('ActivityPub activity'),
      '#default_value' => $this->getConfiguration()['activity'],
      '#options' => $activitypub_activity_options,
    ];

    $form['object'] = [
      '#type' => 'select',
      '#title' => $this->t('ActivityPub object'),
      '#default_value' =>$this->getConfiguration()['object'],
      '#options' => $activitypub_object_options,
    ];

    return $form;
  }

}
