<?php

namespace Drupal\Tests\activitypub\Functional;

use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Url;
use Drupal\Tests\BrowserTestBase;
use Drupal\user\Entity\Role;
use Drupal\user\RoleInterface;
use GuzzleHttp\Exception\RequestException;

abstract class ActivityPubTestBase extends BrowserTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'activitypub',
    'activitypub_test',
    'webfinger',
    'nodeinfo',
    'page_cache',
    'block',
    'node',
    'user',
  ];

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The name for the first account.
   *
   * @var string
   */
  protected $accountNameOne = 'NameOne';

  /**
   * The name for the second account.
   *
   * @var string
   */
  protected $accountNameTwo = 'NameTwo';

  /**
   * The default theme to use.
   *
   * @var string
   */
  protected $defaultTheme = 'stark';

  /**
   * A user with all permissions.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $adminUser;

  /**
   * An authenticated user with less permissions.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $authenticatedUserOne;

  /**
   * An authenticated user with less permissions.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $authenticatedUserTwo;

  /**
   * The authenticated user permissions.
   *
   * @var array
   */
  protected $authenticatedUserPermissions = [
    'allow users to enable activitypub',
    'administer nodes',
    'bypass node access',
    'access user profiles'
  ];

  /**
   * Protected bundles.
   *
   * @var array
   */
  protected $bundles = [];

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();

    /** @var \Drupal\user\RoleInterface $role */
    $role = Role::load(RoleInterface::ANONYMOUS_ID);
    $this->grantPermissions($role, ['access user profiles']);

    try {
      $this->adminUser = $this->createUser([], 'administrator', TRUE);
    } catch (EntityStorageException $ignored) {}
    try {
      $this->authenticatedUserOne = $this->createUser($this->authenticatedUserPermissions, 'fediverseOne');
    } catch (EntityStorageException $ignored) {}
    try {
      $this->authenticatedUserTwo = $this->createUser($this->authenticatedUserPermissions, 'fediverseTwo');
    } catch (EntityStorageException $ignored) {}

    $this->httpClient = $this->container->get('http_client_factory')
      ->fromOptions(['base_uri' => $this->baseUrl]);

  }

  /**
   * Helper method to activate ActivityPub for authenticated user One.
   *
   * @param $assert_session
   * @param bool $enable_second_account
   */
  protected function enableActivityPub($assert_session, $enable_second_account = FALSE) {
    $this->drupalLogin($this->authenticatedUserOne);
    $edit = [
      'activitypub_enable' => TRUE,
    ];

    $this->drupalGet('user/' . $this->authenticatedUserOne->id() . '/activitypub');
    $assert_session->responseContains('ActivityPub is not enabled for your account.');

    $this->drupalPostForm('user/' . $this->authenticatedUserOne->id() . '/activitypub/settings', $edit, 'Save');
    $assert_session->responseContains('Please enter your ActivityPub username.');
    $edit = [
      'activitypub_enable' => TRUE,
      'activitypub_name' => $this->accountNameOne,
    ];
    $this->drupalPostForm('user/' . $this->authenticatedUserOne->id() . '/activitypub/settings', $edit, 'Save');

    if ($enable_second_account) {
      $this->drupalLogin($this->authenticatedUserTwo);
      $edit = [
        'activitypub_enable' => TRUE,
        'activitypub_name' => $this->accountNameTwo,
      ];
      $this->drupalPostForm('user/' . $this->authenticatedUserTwo->id() . '/activitypub/settings', $edit, 'Save');
    }
  }

  /**
   * Create a type config entity.
   *
   * @param $activity
   * @param $bundle
   * @param $object
   * @param $target_entity_type_id
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function createType($activity, $bundle, $object, $target_entity_type_id = 'node') {
    if (!isset($this->bundles[$bundle])) {
      $this->bundles[$bundle] = TRUE;
      $this->drupalCreateContentType(['type' => 'page']);
    }

    /** @var \Drupal\activitypub\Entity\ActivityPubTypeInterface $entity */
    $values = [
      'status' => TRUE,
      'label' => $bundle . ' - ' . $object,
      'id' => 'map',
      'plugin' => [
        'id' => 'activitypub_dynamic_types',
        'configuration' => [
          'activity' => $activity,
          'object' => $object,
          'target_bundle' => $bundle,
          'target_entity_type_id' => $target_entity_type_id,
          'field_mapping' => [
            [
              'field_name' => 'created',
              'property' => 'published',
            ],
            [
              'field_name' => 'body',
              'property' => 'content',
            ],
          ],
        ]
      ],
    ];
    $entity = \Drupal::entityTypeManager()->getStorage('activitypub_type')->create($values);
    $entity->save();
  }

  /**
   * Generate dummy Activity for a user id.
   *
   * @param $uid
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function createActivity($uid) {
    $values = [
      'status' => TRUE,
      'uid' => $uid,
      'collection' => ActivityPubActivityInterface::INBOX,
      'type' => 'Type',
      'actor' => 'actor',
      'object' => 'object',
    ];
    $entity = \Drupal::entityTypeManager()->getStorage('activitypub_activity')->create($values);
    $entity->save();
  }

  /**
   * Get resource url.
   *
   * @param $account
   * @param $addAcctUriScheme
   *
   * @return string.
   */
  protected function getResourceUrl($account, $addAcctUriScheme = TRUE) {
    $base_url = Url::fromRoute('<front>', [], ['absolute' => TRUE])->toString();
    return ($addAcctUriScheme ? 'acct:' : '') . $account . '@' . str_replace(['http://', 'https://'], '', $base_url);
  }

  /**
   * Send inbox request.
   *
   * @param $url
   * @param $payload
   *
   * @return \Psr\Http\Message\ResponseInterface|null
   */
  protected function sendInboxRequest($url, $payload) {

    $headers = [
      'accept' => 'application/activity+json',
    ];

    try {
      $response = $this->httpClient->post($url, ['json' => $payload, 'headers' => $headers]);
    }
    catch (RequestException $e) {
      $response = $e->getResponse();
    }

    return $response;

  }

  /**
   * Runs the inbox queue.
   */
  protected function runInboxQueue() {
    if (\Drupal::config('activitypub.settings')->get('process_inbox_handler') == 'drush') {
      \Drupal::service('activitypub.process.client')->handleInboxQueue();
    }
  }

  /**
   * Runs the outbox queue.
   */
  protected function runOutboxQueue() {
    if (\Drupal::config('activitypub.settings')->get('process_outbox_handler') == 'drush') {
      \Drupal::service('activitypub.process.client')->handleOutboxQueue();
    }
  }

  /**
   * Clear queue.
   *
   * @param $queue_name
   */
  protected function clearQueue($queue_name) {
    \Drupal::database()->delete('queue')->condition('name', $queue_name)->execute();
  }

  /**
   * Flush cache bins.
   *
   * I don't understand why the cache is not cleared in the test because it
   * works fine manually.
   */
  protected function flushBins() {
    $module_handler = \Drupal::moduleHandler();
    $module_handler->invokeAll('cache_flush');
    foreach (Cache::getBins() as $service_id => $cache_backend) {
      $cache_backend->deleteAll();
    }
  }

  /**
   * Follow a user.
   *
   * @param $actor
   * @param $object
   * @param int $actor_nr
   * @param bool $run_queue
   */
  protected function followUser($actor, $object, $actor_nr = 1, $run_queue = FALSE) {
    if ($actor_nr == 2) {
      $this->drupalLogin($this->authenticatedUserTwo);
    }
    else {
      $this->drupalLogin($this->authenticatedUserOne);
    }
    $edit = [
      'config_id' => 'follow',
      'actor' => $actor,
      'object' => $object,
      'type' => 'Follow',
      'collection' => 'outbox',
      'status' => 0,
    ];
    $this->drupalPostForm('activitypub/add', $edit, 'Save');
    $this->drupalLogout();

    if ($run_queue) {
      $this->runOutboxQueue();
    }
  }

  /**
   * View activities of user 1 and 2.
   */
  protected function viewActivityOverview() {
    $this->drupalLogin($this->authenticatedUserOne);
    $this->drupalGet('user/' . $this->authenticatedUserOne->id() . '/activitypub');
    $this->drupalLogout();
    $this->drupalLogin($this->authenticatedUserTwo);
    $this->drupalGet('user/' . $this->authenticatedUserTwo->id() . '/activitypub');
    $this->drupalLogout();
    $this->drupalGet('user/' . $this->authenticatedUserOne->id() . '/activitypub/' . $this->accountNameOne . '/following', ['query' => ['page' => 0]]);
    $this->drupalGet('user/' . $this->authenticatedUserOne->id() . '/activitypub/' . $this->accountNameOne . '/followers', ['query' => ['page' => 0]]);
    $this->drupalGet('user/' . $this->authenticatedUserTwo->id() . '/activitypub/' . $this->accountNameTwo . '/following', ['query' => ['page' => 0]]);
    $this->drupalGet('user/' . $this->authenticatedUserTwo->id() . '/activitypub/' . $this->accountNameTwo . '/followers', ['query' => ['page' => 0]]);
  }

  /**
   * Set the outbox handler.
   */
  protected function setOutboxHandler() {
    $this->drupalLogin($this->adminUser);
    $edit = [
      'process_outbox_handler' => 'drush',
    ];
    $this->drupalPostForm('admin/config/services/activitypub', $edit, 'Save configuration');
    $this->drupalLogout();;
  }

}
