<?php

namespace Drupal\Tests\activitypub\Functional;

use Drupal\Core\Url;
use Drupal\user\Entity\Role;
use Drupal\user\RoleInterface;

/**
 * Tests comment functionality.
 *
 * @group activitypub
 */
class CommentTest extends ActivityPubTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'activitypub_test',
    'activitypub',
    'webfinger',
    'nodeinfo',
    'block',
    'comment',
    'field',
    'node',
    'text',
    'user',
    'field_ui',
  ];

  /**
   * The authenticated user permissions.
   *
   * @var array
   */
  protected $authenticatedUserPermissions = [
    'allow users to enable activitypub',
    'administer nodes',
    'bypass node access',
    'access user profiles',
    'access comments',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(){
    parent::setUp();

    /** @var \Drupal\user\RoleInterface $role */
    $role = Role::load(RoleInterface::ANONYMOUS_ID);
    $this->grantPermissions($role, ['access comments', 'post comments']);
  }

  /**
   * Test comment request.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function testComment() {
    $assert_session = $this->assertSession();

    $this->enableActivityPub($assert_session, TRUE);
    $this->drupalLogout();
    $this->drupalLogin($this->adminUser);
    $edit = [
      'process_inbox_handler' => 'drush',
      'comment_create_enable' => TRUE,
    ];
    $this->drupalPostForm('admin/config/services/activitypub', $edit, 'Save configuration');
    $edit = ['existing_storage_name' => 'activitypub_activity', 'existing_storage_label' => 'Activity reference'];
    $this->drupalPostForm('admin/structure/comment/manage/comment/fields/add-field', $edit, 'Save and continue');
    $this->drupalPostForm('admin/structure/comment/manage/comment/display', ['fields[activitypub_activity][region]' => 'hidden'], 'Save');

    $node_content = 'But who will read it?';
    $node_values = [
      'type' => 'article',
      'uid' => $this->authenticatedUserTwo->id(),
      'title' => 'This is my first post',
      'body' => ['value' => $node_content],
    ];
    $node = $this->drupalCreateNode($node_values);
    $this->drupalLogout();

    // Actor href.
    $actor_href = Url::fromRoute('activitypub.user.self', ['user' => $this->authenticatedUserOne->id(), 'activitypub_actor' => $this->accountNameOne], ['absolute' => TRUE])->toString();

    $payload = [
      'type' => 'Create',
      '@context' => 'https://www.w3.org/ns/activitystreams',
      'id' => $actor_href . '/reply/1',
      'actor' => $actor_href,
      'to' => ['https://www.w3.org/ns/activitystreams#Public'],
      'object' => [
        'type' => 'Note',
        'to' => ['https://www.w3.org/ns/activitystreams#Public'],
        'attributedTo' => $actor_href,
        'published' => '2020-03-12T09:27:03Z',
        'content' => 'I just found it!',
        'inReplyTo' => $node->toUrl('canonical', ['absolute' => TRUE])->toString()
      ],
    ];
    $inbox = Url::fromRoute('activitypub.inbox', ['user' => $this->authenticatedUserTwo->id(), 'activitypub_actor' => $this->accountNameTwo])->toString();
    $response = $this->sendInboxRequest($inbox, $payload);
    self::assertEquals(202, $response->getStatusCode());

    /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorageInterface $activityStorage */
    $activityStorage = \Drupal::entityTypeManager()->getStorage('activitypub_activity');
    /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity */
    $activity = $activityStorage->load(1);
    self::assertEquals("Create", $activity->getType());
    self::assertEquals($payload['id'], $activity->getExternalId());
    self::assertEquals($actor_href, $activity->getActor());
    self::assertEquals($payload['object']['inReplyTo'], $activity->getObject());

    $count = \Drupal::queue(ACTIVITYPUB_INBOX_QUEUE)->numberOfItems();
    self::assertEquals(1, $count);

    $this->runInboxQueue();
    $this->drupalGet('node/1');
    $assert_session->responseContains($node_content);
    $assert_session->responseNotContains('Activity reference');

    // This fails on the test bot, most likely due to the subdirectory.
    /** @var \Drupal\comment\CommentStorageInterface $commentStorage */
    //$commentStorage = \Drupal::entityTypeManager()->getStorage('comment');
    /** @var \Drupal\comment\CommentInterface $comment */
    //$comment = $commentStorage->load(1);
    //self::assertEquals($payload['object']['content'], $comment->get('comment_body')->value);
    //self::assertTrue($comment->isPublished());
    //$assert_session->responseContains($payload['object']['content']);
  }

}
