<?php

namespace Drupal\Tests\activitypub\Functional;

use Drupal\Core\Url;

/**
 * Tests Follow functionality.
 *
 * @group activitypub
 */
class FollowTest extends ActivityPubTestBase {

  /**
   * Test follow functionality.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Behat\Mink\Exception\ExpectationException
   */
  public function testFollow() {
    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Set outbox handler.
    $this->setOutboxHandler();

    // Setup ActivityPub actors.
    $this->enableActivityPub($assert_session, TRUE);
    $this->drupalLogout();
    $actor_href = Url::fromRoute('activitypub.user.self', ['user' => $this->authenticatedUserOne->id(), 'activitypub_actor' => $this->accountNameOne], ['absolute' => TRUE])->toString();
    $object_href = Url::fromRoute('activitypub.user.self', ['user' => $this->authenticatedUserTwo->id(), 'activitypub_actor' => $this->accountNameTwo], ['absolute' => TRUE])->toString();

    $payload = [
      'type' => 'Follow',
      '@context' => 'https://www.w3.org/ns/activitystreams',
      'id' => $actor_href . '/follow/1',
      'actor' => $actor_href,
      'object' => $object_href,
    ];
    $inbox = Url::fromRoute('activitypub.inbox', ['user' => $this->authenticatedUserTwo->id(), 'activitypub_actor' => $this->accountNameTwo])->toString();
    $response = $this->sendInboxRequest($inbox, $payload);
    self::assertEquals(202, $response->getStatusCode());

    /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorageInterface $storage */
    $storage = \Drupal::entityTypeManager()->getStorage('activitypub_activity');
    /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity */
    $activity = $storage->load(1);
    self::assertEquals(FALSE, $activity->isPublished());
    self::assertEquals("Follow", $activity->getType());
    self::assertEquals($payload['id'], $activity->getExternalId());
    self::assertEquals($actor_href, $activity->getActor());
    self::assertEquals($object_href, $activity->getObject());

    $activity = $storage->load(2);
    self::assertEquals(FALSE, $activity->isPublished());
    self::assertEquals("Accept", $activity->getType());
    self::assertEquals($object_href, $activity->getActor());
    self::assertEquals($actor_href, $activity->getObject());
    self::assertEquals($payload['id'], $activity->getExternalId());

    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(1, $count);

    // Account 1 check
    $this->assertFollowersAndFollowing($this->authenticatedUserOne->id(), $this->accountNameOne, 0, 0, '', '', $assert_session, $page);

    // Account 2 check
    $this->assertFollowersAndFollowing($this->authenticatedUserTwo->id(), $this->accountNameTwo, 0, 0, '', '', $assert_session, $page);

    $this->runOutboxQueue();
    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(0, $count);

    $storage->resetCache([1, 2]);
    $activity = $storage->load(1);
    self::assertEquals(TRUE, $activity->isPublished());

    // Account 1 check
    $this->assertFollowersAndFollowing($this->authenticatedUserOne->id(), $this->accountNameOne, 0, 0, '', '', $assert_session, $page);

    // Account 2 check
    $this->assertFollowersAndFollowing($this->authenticatedUserTwo->id(), $this->accountNameTwo, 1, 0, $actor_href, '', $assert_session, $page);

    // Allow to block incoming domains.
    $this->drupalLogin($this->authenticatedUserTwo);
    $edit = [
      'activitypub_blocked_domains' => 'http://blocked.domain*',
    ];
    $this->drupalPostForm('user/' . $this->authenticatedUserTwo->id() . '/activitypub/settings', $edit, 'Save');
    $assert_session->responseContains('http://blocked.domain*');
    $this->drupalLogout();

    $payload = [
      'type' => 'Follow',
      '@context' => 'https://www.w3.org/ns/activitystreams',
      'id' => $actor_href . '/follow/2',
      'actor' => 'http://blocked.domain/user/spammer',
      'object' => $object_href,
    ];
    $response = $this->sendInboxRequest($inbox, $payload);
    self::assertEquals(403, $response->getStatusCode());

    $payload['actor'] = 'http://allowed.domain/user/good';
    $payload['id'] = 'http://allowed.domain/user/good/follow/2';
    $payload['object'] = (object) [
      'object' => $object_href,
    ];
    $response = $this->sendInboxRequest($inbox, $payload);
    self::assertEquals(202, $response->getStatusCode());
    $activity = $storage->load(4);
    self::assertEquals("Follow", $activity->getType());
    self::assertEquals($payload['actor'], $activity->getActor());
    self::assertEquals($object_href, $activity->getObject());

    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(1, $count);

    $activity = $storage->load(5);
    self::assertEquals("outbox", $activity->getCollection());
    self::assertEquals("Accept", $activity->getType());
    self::assertEquals($payload['actor'], $activity->getObject());
    self::assertEquals($object_href, $activity->getActor());
    $this->runOutboxQueue();

    // Account 1 check
    $this->assertFollowersAndFollowing($this->authenticatedUserOne->id(), $this->accountNameOne, 0, 0,'' , '', $assert_session, $page);

    // Account 2 check
    $this->assertFollowersAndFollowing($this->authenticatedUserTwo->id(), $this->accountNameTwo, 2, 0, [$payload['actor'], $actor_href], '', $assert_session, $page);

    // Remove follower (via undo).
    $payload['type'] = 'Undo';
    $payload['id'] = 'http://allowed.domain/user/good/follow/3';
    $payload['actor'] = 'http://allowed.domain/user/good';
    $payload['object'] = (object) [
      'type' => 'Follow',
      'actor' => $payload['actor'],
      'object' => $object_href,
    ];
    $response = $this->sendInboxRequest($inbox, $payload);
    self::assertEquals(202, $response->getStatusCode());

    // Account 1 check
    $this->assertFollowersAndFollowing($this->authenticatedUserOne->id(), $this->accountNameOne, 0, 0, '', '', $assert_session, $page);

    // Account 2 check
    $this->assertFollowersAndFollowing($this->authenticatedUserTwo->id(), $this->accountNameTwo, 1, 0, $actor_href, '', $assert_session, $page);

    $this->drupalGet('nodeinfo/content');
    $content = json_decode($page->getContent());
    self::assertEquals(2, $content->usage->users->total);
    self::assertEquals(0, $content->usage->localPosts);

    // Send a DELETE where actor and object are equal, this activity will not be
    // saved.
    $payload = [
      'type' => 'Delete',
      'id' => 'https://mastodon.social/user/deleted#delete',
      'actor' => 'https://mastodon.social/user/deleted',
      'object' => 'https://mastodon.social/user/deleted',
    ];
    $response = $this->sendInboxRequest($inbox, $payload);
    self::assertEquals(202, $response->getStatusCode());
    $ids = $storage->getQuery()->condition('uid', $this->authenticatedUserTwo->id())->execute();
    self::assertEquals(2, count($ids));

    // Make sure there's only one follower.
    $payload = [
      'type' => 'Follow',
      '@context' => 'https://www.w3.org/ns/activitystreams',
      'id' => $actor_href . '/follow/1',
      'actor' => $actor_href,
      'object' => $object_href,
    ];
    $inbox = Url::fromRoute('activitypub.inbox', ['user' => $this->authenticatedUserTwo->id(), 'activitypub_actor' => $this->accountNameTwo])->toString();
    $response = $this->sendInboxRequest($inbox, $payload);
    self::assertEquals(202, $response->getStatusCode());
    $this->drupalGet('user/' . $this->authenticatedUserTwo->id() . '/activitypub/' . $this->accountNameTwo . '/followers', ['query' => ['page' => 0]]);
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals(1, count($content->orderedItems));
    $ids = $storage->getQuery()->condition('uid', $this->authenticatedUserTwo->id())->execute();
    self::assertEquals(2, count($ids));
    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(0, $count);

    // Test follow started from internal.
    $this->followUser($object_href, $actor_href, 2);

    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(1, $count);
    $activity = $storage->load(6);
    self::assertEquals("outbox", $activity->getCollection());
    self::assertEquals("Follow", $activity->getType());
    self::assertEquals($object_href, $activity->getActor());
    self::assertEquals($actor_href, $activity->getObject());

    // Account 1 check
    $this->assertFollowersAndFollowing($this->authenticatedUserOne->id(), $this->accountNameOne, 0, 0, '', '', $assert_session, $page);

    // Account 2 check
    $this->assertFollowersAndFollowing($this->authenticatedUserTwo->id(), $this->accountNameTwo, 1, 0, $actor_href, '', $assert_session, $page);

    $this->runOutboxQueue();
    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(0, $count);

    $activity = $storage->load(7);
    self::assertEquals("inbox", $activity->getCollection());
    self::assertEquals("Follow", $activity->getType());
    self::assertEquals($object_href, $activity->getActor());
    self::assertEquals($actor_href, $activity->getObject());

    $activity = $storage->load(8);
    self::assertEquals("outbox", $activity->getCollection());
    self::assertEquals("Accept", $activity->getType());
    self::assertEquals($actor_href, $activity->getActor());
    self::assertEquals($object_href, $activity->getObject());

    $activity = $storage->load(9);
    self::assertEquals("inbox", $activity->getCollection());
    self::assertEquals("Accept", $activity->getType());
    self::assertEquals($actor_href, $activity->getActor());
    self::assertEquals($object_href, $activity->getObject());

    // Account 1 check
    $this->assertFollowersAndFollowing($this->authenticatedUserOne->id(), $this->accountNameOne, 1, 0, $object_href, '', $assert_session, $page);

    // Account 2 check
    $this->assertFollowersAndFollowing($this->authenticatedUserTwo->id(), $this->accountNameTwo, 1, 1, $actor_href, $actor_href, $assert_session, $page);

    //$this->viewActivityOverview();

    // Undo follow.
    $this->drupalLogin($this->authenticatedUserTwo);
    $this->drupalGet('user/' . $this->authenticatedUserTwo->id() . '/activitypub');
    $this->drupalGet('activitypub/6/undo');
    $this->drupalLogout();

    $activity = $storage->load(10);
    self::assertEquals("outbox", $activity->getCollection());
    self::assertEquals("Undo", $activity->getType());
    self::assertEquals($object_href, $activity->getActor());
    self::assertEquals($actor_href, $activity->getObject());

    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(1, $count);

    $this->runOutboxQueue();
    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(0, $count);

    //$this->viewActivityOverview();

    $storage->resetCache();
    $activity = $storage->load(6);
    self::assertNull($activity);
    $activity = $storage->load(7);
    self::assertNull($activity);
    $activity = $storage->load(8);
    self::assertNull($activity);
    $activity = $storage->load(9);
    self::assertNull($activity);
    $activity = $storage->load(10);
    self::assertNotNull($activity);
    $activity = $storage->load(11);
    self::assertNull($activity);

    // Account 1 check
    $this->assertFollowersAndFollowing($this->authenticatedUserOne->id(), $this->accountNameOne, 0, 0, '', '', $assert_session, $page);

    // Account 2 check
    $this->assertFollowersAndFollowing($this->authenticatedUserTwo->id(), $this->accountNameTwo, 1, 0, $actor_href, '', $assert_session, $page);
  }

  /**
   * Assert followers and following.
   *
   * @param $id
   * @param $name
   * @param $followers
   * @param $following
   * @param $followers_urls
   * @param $following_urls
   * @param $assert_session
   * @param $page
   */
  protected function assertFollowersAndFollowing($id, $name, $followers, $following, $followers_urls, $following_urls, $assert_session, $page) {
    $this->drupalGet('user/' . $id . '/activitypub/' . $name . '/followers');
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals($followers, $content->totalItems);

    $this->drupalGet('user/' . $id . '/activitypub/' . $name . '/followers', ['query' => ['page' => 0]]);
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals($followers, $content->totalItems);
    self::assertEquals($followers, count($content->orderedItems));

    if (is_array($followers_urls)) {
      foreach ($followers_urls as $u) {
        self::assertTrue(in_array($u, $content->orderedItems));
      }
    }
    elseif (!empty($followers_urls)) {
      self::assertTrue(in_array($followers_urls, $content->orderedItems));
    }

    $this->drupalGet('user/' . $id . '/activitypub/' . $name . '/following');
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals($following, $content->totalItems);

    $this->drupalGet('user/' . $id . '/activitypub/' . $name . '/following', ['query' => ['page' => 0]]);
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals($following, $content->totalItems);
    self::assertEquals($following, count($content->orderedItems));

    if (is_array($following_urls)) {
      foreach ($following_urls as $u) {
        self::assertTrue(in_array($u, $content->orderedItems));
      }
    }
    elseif (!empty($following_urls)) {
      self::assertTrue(in_array($following_urls, $content->orderedItems));
    }
  }

}
