<?php

namespace Drupal\Tests\activitypub\Functional;

use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\Core\Url;

/**
 * Tests Outbox functionality.
 *
 * @group activitypub
 */
class OutboxTest extends ActivityPubTestBase {

  /**
   * Test outbox responses.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function testOutboxResponse() {
    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    $this->enableActivityPub($assert_session);
    $this->drupalLogout();

    $outbox = Url::fromRoute('activitypub.outbox', ['user' => $this->authenticatedUserOne->id(), 'activitypub_actor' => $this->accountNameOne], ['absolute' => TRUE])->toString();
    $this->drupalGet($outbox);
    $assert_session->statusCodeEquals(200);

    $content = json_decode($page->getContent());
    self::assertEquals("OrderedCollection", $content->type);
    $first_page = $content->first;
    $this->drupalGet($content->first);
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals($first_page, $content->id);

    // Create type.
    $this->createType('Create', 'page', 'Note');

    // Send to outbox.
    $this->drupalLogin($this->authenticatedUserOne);
    $this->drupalGet('node/add/page');
    $assert_session->statusCodeEquals(200);
    $assert_session->pageTextContains('ActivityPub outbox');

    $edit = [
      'title[0][value]' => 'Page one',
      'body[0][value]' => 'Hello world!',
      'activitypub_type' => 'map',
    ];
    $this->drupalPostForm(NULL, $edit, 'Save');
    $assert_session->addressEquals('node/1');
    $this->drupalGet('node/1/edit');
    $assert_session->pageTextContains('This post has been sent to the outbox.');
    $this->drupalLogout();

    $this->drupalGet($outbox);
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals(1, $content->totalItems);
    $this->drupalGet($content->first);
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals(1, count($content->orderedItems));

    $this->drupalGet('nodeinfo/content');
    $content = json_decode($page->getContent());
    self::assertEquals(1, $content->usage->users->total);
    self::assertEquals(1, $content->usage->localPosts);

    $this->drupalLogin($this->authenticatedUserOne);
    $this->drupalGet('node/add/page');
    $edit = [
      'title[0][value]' => 'Page two',
      'body[0][value]' => 'Hello world!',
      'activitypub_type' => 'map',
    ];
    $this->drupalPostForm(NULL, $edit, 'Save');
    $assert_session->addressEquals('node/2');
    $this->drupalGet('node/2/edit');
    $assert_session->pageTextContains('This post has been sent to the outbox.');
    $this->drupalLogout();

    $this->drupalGet($outbox);
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals(2, $content->totalItems);
    $this->drupalGet($content->first);
    $assert_session->statusCodeEquals(200);
    $content = json_decode($page->getContent());
    self::assertEquals(2, count($content->orderedItems));

    $this->drupalGet('nodeinfo/content', ['query' => ['page' => 1]]);
    $content = json_decode($page->getContent());
    self::assertEquals(1, $content->usage->users->total);
    self::assertEquals(2, $content->usage->localPosts);
  }

  /**
   * Test to functionality.
   */
  public function testTo() {
    $assert_session = $this->assertSession();

    /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorageInterface $storage */
    $storage = \Drupal::entityTypeManager()->getStorage('activitypub_activity');

    // Setup ActivityPub actors.
    $this->enableActivityPub($assert_session, TRUE);
    $this->drupalLogout();

    // Set outbox handler.
    $this->setOutboxHandler();

    $actor_href = Url::fromRoute('activitypub.user.self', ['user' => $this->authenticatedUserOne->id(), 'activitypub_actor' => $this->accountNameOne], ['absolute' => TRUE])->toString();
    $object_href = Url::fromRoute('activitypub.user.self', ['user' => $this->authenticatedUserTwo->id(), 'activitypub_actor' => $this->accountNameTwo], ['absolute' => TRUE])->toString();

    // Create type.
    $this->createType('Create', 'page', 'Note');

    // Send to outbox.
    $this->drupalLogin($this->authenticatedUserOne);
    $this->drupalGet('node/add/page');
    $edit = [
      'title[0][value]' => 'Sending to someone',
      'body[0][value]' => 'Hello unknown follower!',
      'activitypub_type' => 'map',
      'activitypub_to' => $object_href,
    ];
    $this->drupalPostForm(NULL, $edit, 'Save');
    $this->drupalLogout();

    /** @var \Drupal\node\NodeInterface $node */
    $node = \Drupal::entityTypeManager()->getStorage('node')->load(1);

    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(1, $count);
    $this->runOutboxQueue();
    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(0, $count);

    //$this->viewActivityOverview();

    /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity */
    $activity = $storage->load(2);
    self::assertEquals($actor_href, $activity->getActor());
    self::assertEquals(ActivityPubActivityInterface::INBOX, $activity->getCollection());
    self::assertEquals('Create', $activity->getType());
    self::assertEquals($this->authenticatedUserTwo->id(), $activity->getOwnerId());
    self::assertEquals($node->toUrl('canonical', ['absolute' => TRUE])->toString(), $activity->getObject());
    $json = json_decode($activity->getPayLoad(), TRUE);
    self::assertEquals('Note', $json['object']['type']);
    self::assertEquals($json['id'], $activity->getExternalId());
    self::assertEquals($node->toUrl('canonical', ['absolute' => TRUE])->toString(), $json['object']['id']);
    self::assertEquals($node->get('body')->value, $json['object']['content']);
  }

  /**
   * Tests delete requests.
   */
  public function testDeleteRequest() {
    $assert_session = $this->assertSession();

    /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorageInterface $storage */
    $storage = \Drupal::entityTypeManager()->getStorage('activitypub_activity');

    // Setup ActivityPub actors.
    $this->enableActivityPub($assert_session, TRUE);
    $this->drupalLogout();

    // Set outbox handler.
    $this->setOutboxHandler();

    $actor_href = Url::fromRoute('activitypub.user.self', ['user' => $this->authenticatedUserOne->id(), 'activitypub_actor' => $this->accountNameOne], ['absolute' => TRUE])->toString();
    $object_href = Url::fromRoute('activitypub.user.self', ['user' => $this->authenticatedUserTwo->id(), 'activitypub_actor' => $this->accountNameTwo], ['absolute' => TRUE])->toString();
    $this->followUser($object_href, $actor_href, 2, TRUE);
    //$this->viewActivityOverview();

    // Create type.
    $this->createType('Create', 'page', 'Note');

    // Send to outbox.
    $this->drupalLogin($this->authenticatedUserOne);
    $this->drupalGet('node/add/page');
    $edit = [
      'title[0][value]' => 'Page one',
      'body[0][value]' => 'Hello world!',
      'activitypub_type' => 'map',
    ];
    $this->drupalPostForm(NULL, $edit, 'Save');
    $this->drupalLogout();

    /** @var \Drupal\node\NodeInterface $node */
    $node = \Drupal::entityTypeManager()->getStorage('node')->load(1);

    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(1, $count);
    $this->runOutboxQueue();
    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(0, $count);

    //$this->viewActivityOverview();

    /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity */
    $activityCreate = $storage->load(5);
    $activityCreateUrl = $activityCreate->toUrl('canonical', ['absolute' => TRUE])->toString();
    $activity = $storage->load(6);
    $nodeUrl = $node->toUrl('canonical', ['absolute' => TRUE])->toString();
    self::assertEquals(ActivityPubActivityInterface::INBOX, $activity->getCollection());
    self::assertEquals('Create', $activity->getType());
    self::assertEquals($this->authenticatedUserTwo->id(), $activity->getOwnerId());
    self::assertEquals($nodeUrl, $activity->getObject());
    $json = json_decode($activity->getPayLoad(), TRUE);
    self::assertEquals('Note', $json['object']['type']);
    self::assertEquals($json['id'], $activity->getExternalId());
    self::assertEquals($node->toUrl('canonical', ['absolute' => TRUE])->toString(), $json['object']['id']);
    self::assertEquals($node->get('body')->value, $json['object']['content']);

    $this->drupalLogin($this->authenticatedUserOne);
    $this->drupalPostForm('node/' . $node->id() . '/delete', [], 'Delete');
    $count = \Drupal::queue(ACTIVITYPUB_OUTBOX_QUEUE)->numberOfItems();
    self::assertEquals(1, $count);

    //$this->viewActivityOverview();
    $this->runOutboxQueue();
    //$this->viewActivityOverview();

    $storage->resetCache([5, 6]);
    $activity = $storage->load(5);
    self::assertNull($activity);
    $activity = $storage->load(6);
    self::assertNull($activity);
    $activity = $storage->load(7);
    self::assertEquals('Delete', $activity->getType());
    self::assertEquals($actor_href, $activity->getActor());
    self::assertEquals($nodeUrl, $activity->getObject());
    self::assertEquals($activityCreateUrl, $activity->getExternalId());
    $activity = $storage->load(8);
    self::assertNull($activity);
  }

}
